/*
 * This file is part of Rahmenwerk.
 *
 * Rahmenwerk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License  as 
 * published by the Free Software Foundation, either version 3 of the
 * License.
 *
 * Rahmenwerk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Rahmenwerk.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 *
 * Copyright (c) 2008 Dambach Werke GmbH.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 * Dambach Werke GmbH - initial API and implementation
 *
 *
 *
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */
package de.bsvrz.buv.rw.basislib.kalender;

import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;

import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Diese Klasse stellt die interne Datenstruktur eines Zeitfilters dar und kann
 * zur Parametrierung und als Rckgabewert des Zeitfilters benutzt werden.
 * Erfasst werden unter anderem die selektierten Monate, Tage und Wochentage
 * sowie eine Liste der zugehrigen Zeitgruppen.
 * 
 * @author HofmannJ
 * 
 */
public class KalenderZeitFilterDatenStruktur implements Cloneable
{

	/**
	 * Der Debugger
	 */
	private Debug _debug = Debug.getLogger();

	private static int _anzahlMonate = 12;

	private static int _anzahlTage = 31;

	private static int _anzahlWochentage = 7;

	@SuppressWarnings("unused")
	private long _anfang24H = -1;
	@SuppressWarnings("unused")
	private long _ende24H = -1;

	/**
	 * Array ber 12 Monate von Januar - Dezember true == Monat ist gewhlt, false ==
	 * Monat ist nicht gewhlt
	 */
	private boolean[] _Monate = new boolean[_anzahlMonate];

	/**
	 * Array ber 31 Tage von 1 - 31 true == Tag ist gewhlt, false == Tag ist
	 * nicht gewhlt
	 */
	private boolean[] _Tage = new boolean[_anzahlTage];

	/**
	 * Array ber 7 Wochentage, Montag bis Sonntag true == Wochentag ist gewhlt,
	 * false == Wochentag ist nicht gewhlt
	 */
	private boolean[] _Wochentage = new boolean[_anzahlWochentage];

	private List<KalenderZeitFilterDatenStrukturZeitGruppe> _listeZeitGruppen;

	/**
	 * Basiskonstruktor dieser Klasse, startet die basisInitialisierung.
	 * 
	 */
	public KalenderZeitFilterDatenStruktur()
	{
		basisInitialisierung();
	}

	/**
	 * Basiskonstruktor dieser Klasse, startet die basisInitialisierung.
	 * 
	 * @param String
	 *          Reprsentation des Filters.
	 */
	public KalenderZeitFilterDatenStruktur(String filter)
	{
		basisInitialisierung();
		erzeugeAusFilterString(filter);
	}

	/**
	 * Fhrt eine Basisinitialisierung der Felder Monate, Tage, Wochentage mit
	 * false durch
	 * 
	 */
	private void basisInitialisierung()
	{
		GregorianCalendar anfang = new GregorianCalendar(1970, 1, 1, 0, 0, 0);
		_anfang24H = anfang.getTimeInMillis();
		GregorianCalendar ende = new GregorianCalendar(1970, 1, 2, 0, 0, 0);
		ende.add(Calendar.MILLISECOND, -1);

		_ende24H = ende.getTimeInMillis();

		// Monate auf False setzen
		for (int i = 0; i < _anzahlMonate; i++)
		{
			_Monate[i] = false;
		}

		// Tage aus False setzen
		for (int i = 0; i < _anzahlTage; i++)
		{
			_Tage[i] = false;
		}

		// Wochentage aus False setzen
		for (int i = 0; i < _anzahlWochentage; i++)
		{
			_Wochentage[i] = false;
		}

		// Zeitgruppenliste initialisieren
		_listeZeitGruppen = new LinkedList<KalenderZeitFilterDatenStrukturZeitGruppe>();
	}

	/**
	 * Liest ein Array von selektierten Monaten (jeweils als Zahl) ein und setzt
	 * die entsprechenden Positionen im Feld Monate auf true.
	 * 
	 * @param monate
	 *          Array mit den Monaten, die auf true gesetzt werden sollen
	 */
	@SuppressWarnings("unused")
	private void leseSelektierteMonateEin(int[] monate)
	{
		for (int monat : monate)
		{
			if (monat < _anzahlMonate && _Monate[monat] == false)
			{
				_Monate[monat] = true;
			}
		}
	}

	/**
	 * Liest ein Array von selektierten Tagen (jeweils als Zahl) ein und setzt die
	 * entsprechenden Positionen im Feld Tage auf true
	 * 
	 * @param tage
	 *          Array mit den Tagen, die auf true gesetzt werden sollen
	 */
	@SuppressWarnings("unused")
	private void leseSelektierteTageEin(int[] tage)
	{
		for (int tag : tage)
		{
			if (tag < _anzahlTage && _Tage[tag] == false)
			{
				_Tage[tag] = true;
			}
		}
	}

	/**
	 * Liest ein Array von selektierten Wochentagen (jeweils als Zahl) ein und
	 * setzt die entsprechenden Positionen im Feld Wochentage auf true
	 * 
	 * @param wochentage
	 *          Array mit den Wochentagen, die auf true gesetzt werden sollen
	 */
	@SuppressWarnings("unused")
	private void leseSelektierteWochentageEin(int[] wochentage)
	{
		for (int wochentag : wochentage)
		{
			if (wochentag <= _anzahlWochentage && _Wochentage[wochentag - 1] == false)
			{
				_Wochentage[wochentag - 1] = true;
			}
		}

	}

	/**
	 * Liefert ein Array mit den Nummern der selektierten Monate
	 * 
	 * @return Array der selektierten Monate
	 */
	public int[] getSelektierteMonate()
	{

		int[] returnWert = null;
		List<Integer> monate = new LinkedList<Integer>();

		for (int i = 0; i < _Monate.length; i++)
		{
			if (_Monate[i])
			{
				monate.add(Integer.valueOf(i));
			}
		}

		returnWert = new int[monate.size()];
		int j = 0;

		ListIterator<Integer> li = monate.listIterator();
		while (li.hasNext())
		{
			returnWert[j] = li.next().intValue();
			j++;
		}

		return returnWert;
	}

	/**
	 * Liefert ein Array mit den Nummern der selektierten Tage
	 * 
	 * @return Array mit den Nummern der selektierten Tage
	 */
	public int[] getSelektierteTage()
	{

		int[] returnWert = null;
		List<Integer> tage = new LinkedList<Integer>();

		for (int i = 0; i < _Tage.length; i++)
		{
			if (_Tage[i])
			{
				tage.add(Integer.valueOf(i));
			}
		}

		returnWert = new int[tage.size()];
		int j = 0;

		ListIterator<Integer> li = tage.listIterator();
		while (li.hasNext())
		{
			returnWert[j] = li.next().intValue();
			j++;
		}

		return returnWert;
	}

	/**
	 * Liefert ein Array mit den Nummern der selektierten Wochentage.
	 * 
	 * @return Array mit den Nummern der Selektierten Wochentage.
	 */
	public int[] getSelektierteWochentage()
	{

		int[] returnWert = null;
		List<Integer> wochentage = new LinkedList<Integer>();

		for (int i = 0; i < _Wochentage.length; i++)
		{
			if (_Wochentage[i])
			{
				wochentage.add(Integer.valueOf(i));
			}
		}

		// returnWert = new int[wochentage.size()];
		// int j = 0;

		// ListIterator<Integer> li = wochentage.listIterator();
		// while (li.hasNext())
		// {
		// returnWert[j] = li.next().intValue();
		// j++;
		// }

		List<Integer> tempWochentage = new LinkedList<Integer>();
		int i = -1;
		for (Boolean bool : _Wochentage)
		{
			i++;
			if (bool)
			{
				tempWochentage.add(i);
			}
		}

		returnWert = new int[tempWochentage.size()];
		Iterator<Integer> it = tempWochentage.iterator();
		int j = -1;
		while (it.hasNext())
		{
			j++;
			returnWert[j] = it.next();
		}

		return returnWert;
	}

	/**
	 * TextFilter
	 */
	public String erzeugeFilterString()
	{
		String returnString = "";

		String monate = "";
		String tage = "";
		String wochentage = "";
		String zeitgruppen = "";

		/*
		 * Monate
		 */
		int[] selektierteMonate = getSelektierteMonate();
		if (selektierteMonate.length == 0
				|| selektierteMonate.length >= _anzahlMonate)
		{
			monate = "* ";
		} else
		{
			for (int i : selektierteMonate)
			{
				int wert = i + 1;
				monate = monate + wert + ",";
			}
			monate = monate.substring(0, monate.length() - 1) + " ";
		}

		/*
		 * Tage
		 */
		int[] selektierteTage = getSelektierteTage();
		if (selektierteTage.length == 0 || selektierteTage.length >= _anzahlTage)
		{
			tage = "* ";
		} else
		{
			for (int i : selektierteTage)
			{
				int wert = i;
				wert = wert + 1;
				tage = tage + wert + ",";
			}
			tage = tage.substring(0, tage.length() - 1) + " ";
		}

		/*
		 * Wochentage
		 */
		int[] selektierteWochentage = getSelektierteWochentage();
		if (selektierteWochentage.length == 0
				|| selektierteWochentage.length >= _anzahlWochentage)
		{
			wochentage = "* ";
		} else
		{
			for (int i : selektierteWochentage)
			{
				switch (i)
				{
				case 0:
					wochentage = wochentage + "Mo,";
					break;
				case 1:
					wochentage = wochentage + "Di,";
					break;
				case 2:
					wochentage = wochentage + "Mi,";
					break;

				case 3:
					wochentage = wochentage + "Do,";
					break;
				case 4:
					wochentage = wochentage + "Fr,";
					break;
				case 5:
					wochentage = wochentage + "Sa,";
					break;
				case 6:
					wochentage = wochentage + "So,";

				}
			}
			wochentage = wochentage.substring(0, wochentage.length() - 1) + " ";
		}

		/*
		 * Zeitgruppen
		 */
		List<KalenderZeitFilterDatenStrukturZeitGruppe> gruppen = getListeZeitGruppen();
		if (gruppen.size() == 0)
		{
			zeitgruppen = "*";
		} else
		{
			Iterator<KalenderZeitFilterDatenStrukturZeitGruppe> it = gruppen
					.iterator();
			/*
			 * DateFormat df = new SimpleDateFormat("hh:mm:s.S"); String ausgabe =
			 * df.format(date);
			 */
			while (it.hasNext())
			{
				KalenderZeitFilterDatenStrukturZeitGruppe zeitGruppe = it.next();
				long startzeitpunkt = zeitGruppe.getStartzeitpunkt();
				long endzeitpunkt = zeitGruppe.getEndzeitpunkt();

				DateFormat df = new SimpleDateFormat("HH:mm:s.S");
				String startString = df.format(new Date(startzeitpunkt));
				String endeString = df.format(new Date(endzeitpunkt));
				zeitgruppen = zeitgruppen + startString + "-" + endeString + ",";
			}
			zeitgruppen = zeitgruppen.substring(0, zeitgruppen.length() - 1);
		}

		returnString = returnString + monate + tage + wochentage + zeitgruppen;

		return returnString;
	}

	public void erzeugeAusFilterString(String eingang)
	{
		// String eingang = "* 2,4,5 Di 00:00:0.0-00:00:12.0,00:13:0.0-00:44:0.0";
		String[] splitLeer = eingang.split(" ");

		String temp = "===" + getClass().getName() + "::erzeugeAusFilterString===";
		_debug.fine(temp);

		for (String string : splitLeer)
		{
			_debug.fine("Split nach Leerzeichen: " + string);
		}
		String temp2 = "";
		for (int i = 0; i < temp.length(); i++)
		{
			temp2 = temp2 + "=";
		}
		_debug.fine(temp2);

		String monat = splitLeer[0];
		String tage = splitLeer[1];
		String wochentage = splitLeer[2];
		String zeiten = splitLeer[3];
		
		@SuppressWarnings("unused")
		int[] monateRueckgabe = null;
		@SuppressWarnings("unused")
		int[] tageRueckgabe = null;
		@SuppressWarnings("unused")
		int[] wochentageRueckgabe = null;

		/*
		 * Monate
		 */
		if (monat.contains("*"))
		{
			_debug.fine(getClass().getName()
					+ "::erzeugeAusFilterString => Monat komplett");
			
			for(int i = 0; i < _Monate.length; i++)
			{
				_Monate[i] = true;
			}
			
			
		} else
		{
			// Splitten nach dem ,
			String[] monateZahlen = monat.split(",");

			for (String stringMonat : monateZahlen)
			{
				if (stringMonat.contains("-"))
				{
					_debug.fine("-");

					String[] startEndeMonat = stringMonat.split("-");
					int start = Integer.parseInt(startEndeMonat[0]);
					int ende = Integer.parseInt(startEndeMonat[1]);

					_debug.fine("START: " + start + " ENDE: " + ende);

					if ((start > 0) && (ende <= _anzahlMonate))
					{
						for (int i = start - 1; i < ende; i++)
						{
							selektiereMonat(i);
						}
					} else
					{
						_debug
								.error(getClass().getName()
										+ "::erzeugeAusFilterString \n\t=>Der Angegebene Monatsbereich ist ungltig!");
					}

				} else
				{
					_debug.fine("else");

					int tempMonat = Integer.parseInt(stringMonat);
					if (tempMonat <= 12)
					{
						if (tempMonat > 0)
						{
							selektiereMonat(tempMonat - 1);
						} else
						{
							_debug
									.error(getClass().getName()
											+ "::erzeugeAusFilterString \n\t=>Es wurde ein Monat kleiner gleich 0 angegeben. Dieser existiert nicht!");
						}

					} else
					{
						_debug
								.error(getClass().getName()
										+ "::erzeugeAusFilterString \n\t=>Es wurde ein Monat grer 12 angegeben. Dieser existiert nicht!");
					}
				}
			}

			/*
			 * monateRueckgabe = new int[monateZahlen.length]; int i = -1; for (String
			 * string : monateZahlen) { i++; monateRueckgabe[i] =
			 * Integer.parseInt(string); }
			 * 
			 * leseSelektierteMonateEin(monateRueckgabe);
			 */
		}

		/*
		 * Tage
		 */
		if (tage.contains("*"))
		{
			_debug.fine(getClass().getName()
					+ "::erzeugeAusFilterString => Tage komplett");
			
			for(int i = 0; i < _Tage.length; i++)
			{
				_Tage[i] = true;
			}
			
		} else
		{
			String[] stringsTage = tage.split(",");
			for (String stringTag : stringsTage)
			{
				if (stringTag.contains("-"))
				{
					String[] startEndeTage = stringTag.split("-");
					int start = Integer.parseInt(startEndeTage[0]);
					int ende = Integer.parseInt(startEndeTage[1]);
					if ((start > 0) && (ende <= _anzahlTage))
					{
						for (int i = start - 1; i < ende; i++)
						{
							selektiereTag(i);
						}
					}
				} else
				{
					int tag = Integer.parseInt(stringTag);
					tag = tag - 1;
					selektiereTag(tag);
				}
			}

		}

		/*
		 * Wochentage
		 */
		if (wochentage.contains("*"))
		{
			_debug.fine(getClass().getName()
					+ "::erzeugeAusFilterString => Wochentage komplett");
			
			for(int i = 0; i < _Wochentage.length; i++)
			{
				_Wochentage[i] = true;
			}
			
			
		} else
		{
			String[] stringsWochentage = wochentage.split(",");

			for (String string : stringsWochentage)
			{

				if (string.equals("Mo"))
				{
					selektiereWochentag(0);
				}
				if (string.equals("Di"))
				{
					selektiereWochentag(1);
				}
				if (string.equals("Mi"))
				{
					selektiereWochentag(2);
				}
				if (string.equals("Do"))
				{
					selektiereWochentag(3);
				}
				if (string.equals("Fr"))
				{
					selektiereWochentag(4);
				}
				if (string.equals("Sa"))
				{
					selektiereWochentag(5);
				}
				if (string.equals("So"))
				{
					selektiereWochentag(6);
				}

			}

		}

		/*
		 * Zeiten
		 */
		if (zeiten.contains("*"))
		{
			_debug.fine(getClass().getName()
					+ "::erzeugeAusFilterString => Zeiten komplett");

			long zeitVonLong = 2674800000L;
			long zeitBisLong = 2761199999L;
			fuegeZeitGruppeHinzu(zeitVonLong, zeitBisLong);
		} else
		{
			// Splitten nach ,
			String[] stringsZeiten = zeiten.split(",");

			for (String stringZeitKomplett : stringsZeiten)
			{
				if (stringZeitKomplett.contains("-"))
				{
					String[] stringsZeitenBis = stringZeitKomplett.split("-");

					String zeitVon = stringsZeitenBis[0];
					String zeitBis = stringsZeitenBis[1];

					long zeitVonLong = -3600000l;
					long zeitBisLong = -3600000l;

					// Bestimmen Anzahl : in zeitVon
					int anzDoppelpkteZeitVon = zeitVon.split(":").length - 1;
					int anzSemikolaZeitVon = zeitVon.split(".").length - 1;
					int anzDoppelpkteZeitBis = zeitBis.split(":").length - 1;
					int anzSemikolaZeitBis = zeitBis.split(".").length - 1;

					_debug.fine("VONDOPPEL: " + anzDoppelpkteZeitVon + " VONSEMI: "
							+ anzSemikolaZeitVon);
					_debug.fine("BISDOPPEL: " + anzDoppelpkteZeitBis + " BISSEMI: "
							+ anzSemikolaZeitBis);

					// VonZeit
					if (anzDoppelpkteZeitVon == 0)
					{
						DateFormat df = new SimpleDateFormat("HH");
						try
						{
							zeitVonLong = df.parse(zeitVon).getTime();
						} catch (ParseException e)
						{
							e.printStackTrace();
						}
					} else if (anzDoppelpkteZeitVon == 1)
					{
						DateFormat df = new SimpleDateFormat("HH:mm");
						try
						{
							zeitVonLong = df.parse(zeitVon).getTime();
						} catch (ParseException e)
						{
							e.printStackTrace();
						}
					} else if (anzDoppelpkteZeitVon == 2)
					{
						if (anzSemikolaZeitVon == 1)
						{
							DateFormat df = new SimpleDateFormat("HH:mm:s.S");
							try
							{
								zeitVonLong = df.parse(zeitVon).getTime();
							} catch (ParseException e)
							{
								e.printStackTrace();
							}
						} else
						{

							DateFormat df = new SimpleDateFormat("HH:mm:s");
							try
							{
								zeitVonLong = df.parse(zeitVon).getTime();
							} catch (ParseException e)
							{
								e.printStackTrace();
							}
						}
					}

					// Bis-Zeit
					if (anzDoppelpkteZeitBis == 0)
					{
						DateFormat df = new SimpleDateFormat("HH");
						try
						{
							zeitBisLong = df.parse(zeitBis).getTime();
						} catch (ParseException e)
						{
							e.printStackTrace();
						}
					} else if (anzDoppelpkteZeitBis == 1)
					{
						DateFormat df = new SimpleDateFormat("HH:mm");
						try
						{
							zeitBisLong = df.parse(zeitBis).getTime();
						} catch (ParseException e)
						{
							e.printStackTrace();
						}
					} else if (anzDoppelpkteZeitBis == 2)
					{
						if (anzSemikolaZeitBis == 1)
						{
							DateFormat df = new SimpleDateFormat("HH:mm:s.S");
							try
							{
								zeitBisLong = df.parse(zeitBis).getTime();
							} catch (ParseException e)
							{
								e.printStackTrace();
							}
						} else
						{

							DateFormat df = new SimpleDateFormat("HH:mm:s");
							try
							{
								zeitBisLong = df.parse(zeitBis).getTime();
							} catch (ParseException e)
							{
								e.printStackTrace();
							}
						}
					}

					_debug.fine("====> VON: " + zeitVonLong + " BIS: " + zeitBisLong);

					if ((zeitVonLong > -3600001) && (zeitBisLong > -3600001))
					{
						fuegeZeitGruppeHinzu(zeitVonLong, zeitBisLong);
					}

				} else
				// einzelne Zahl, steht fr ganze Stunde
				{
					String zeitVon = stringZeitKomplett;
					long zeitVonLong = -1l;
					long zeitBisLong = -1l;
					DateFormat df = new SimpleDateFormat("HH");
					try
					{
						zeitVonLong = df.parse(zeitVon).getTime();
						zeitBisLong = zeitVonLong + (1000 * 60 * 60) - 1;
						fuegeZeitGruppeHinzu(zeitVonLong, zeitBisLong);
					} catch (ParseException e)
					{
						_debug
								.error(getClass().getName()
										+ "::erzeugeAusFilterString \n\t=>VonZeit bei einzeln angegebener Stunde konnte nicht bestimmt werden.");
						e.printStackTrace();
					}
				}

			}
		
		}

	}

	/*
	 * Selektoren/Deselektoren fr Monate, Tage und Wochentage
	 */

	/**
	 * Selektiert einen Monat und setzt ihn auf true
	 * 
	 * @param i
	 *          Nummer des Monats, der selektiert werden soll
	 */
	public void selektiereMonat(int i)
	{
		_Monate[i] = true;
		_debug.fine(getClass().getName() + "::selektiereMonat \n\t=>Monat " + i
				+ " wurde selektiert.");
	}

	/**
	 * Deselektiert einen Monat und setzt ihn auf false
	 * 
	 * @param i
	 *          Nummer des Monats, der deselektiert soll
	 */
	public void deSelektiereMonat(int i)
	{
		_Monate[i] = false;
		_debug.fine(getClass().getName() + "::deSelektiereMonat \n\t=>Monat " + i
				+ " wurde deselektiert.");

	}

	/**
	 * Selektiert einen Tag und setzt ihn auf true
	 * 
	 * @param i
	 *          Nummer des Tages, der selektiert werden soll
	 */
	public void selektiereTag(int i)
	{
		_Tage[i] = true;
		_debug.fine(getClass().getName() + "::selektiereTag \n\t=>Tag " + i
				+ " wurde selektiert.");

	}

	/**
	 * Deselektiert einen Tag und setzt ihn auf false
	 * 
	 * @param i
	 *          Nummer des tages, der selektiert werden soll
	 * 
	 */
	public void deSelektiereTag(int i)
	{
		_Tage[i] = false;
		_debug.fine(getClass().getName() + "::deSelektiereTag \n\t=>Tag " + i
				+ " wurde deselektiert.");

	}

	/**
	 * Selektiert einen Wochentag und setzt ihn auf true
	 * 
	 * @param i
	 *          Nummer des Wochentags, der selektiert werden soll
	 */
	public void selektiereWochentag(int i)
	{
		_Wochentage[i] = true;
		_debug.fine(getClass().getName() + "::selektiereWochentag \n\t=>Wochentag "
				+ i + " wurde selektiert.");

	}

	/**
	 * Deselektiert einen Wochentag und setzt ihn auf false
	 * 
	 * @param i
	 *          Nummer des Wochentags, der deselektiert werden soll
	 */
	public void deSelektiereWochentag(int i)
	{
		_Wochentage[i] = false;
		_debug.fine(getClass().getName()
				+ "::deSelektiereWochentag \n\t=>Wochentag " + i
				+ " wurde deselektiert.");

	}

	/**
	 * Fgt eine Zeitgruppe zur Liste der Zeitgruppen hinzu
	 * 
	 * @param startzeitpunkt
	 *          Startzeitpunkt als long (Zeit in MS seit 1.1.1970)
	 * @param endzeitpunkt
	 *          Enzeitpunkt als long (Zeit in MS seit 1.1.1970)
	 */
	public void fuegeZeitGruppeHinzu(long startzeitpunkt, long endzeitpunkt)
	{
		KalenderZeitFilterDatenStrukturZeitGruppe zg = new KalenderZeitFilterDatenStrukturZeitGruppe(
				startzeitpunkt, endzeitpunkt);
		_listeZeitGruppen.add(zg);
	}

	/**
	 * Fgt eine Zeitgruppe zur Liste der Zeitgruppen hinzu
	 * 
	 * @param zg
	 *          Zeitgruppe, die hinzugefgt werden soll.
	 */
	public void fuegeZeitGruppeHinzu(KalenderZeitFilterDatenStrukturZeitGruppe zg)
	{
		if (!_listeZeitGruppen.contains(zg))
		{
			_listeZeitGruppen.add(zg);
		} else
		{
			System.err
					.println(getClass().getName()
							+ "::fuegeZeitGruppeHinzu \n\t=>Die bergebene Zeitgruppe ist bereits in der Liste der Zeitgruppen vorhanden.");
		}
	}

	/**
	 * Entfernt eine Zeitgruppe aus der Liste der Zeitgruppen
	 * 
	 * @param zg
	 *          Zeitgruppe, die entfernt werden soll.
	 */
	public void entferneZeitGruppe(KalenderZeitFilterDatenStrukturZeitGruppe zg)
	{
		_listeZeitGruppen.remove(zg);
	}

	/**
	 * Liefert die Liste der ZeitGruppen zurck
	 * 
	 * @return Liste der Zeitgruppen
	 */
	public List<KalenderZeitFilterDatenStrukturZeitGruppe> getListeZeitGruppen()
	{
		return _listeZeitGruppen;
	}

	/**
	 * Setzt die Liste der ZeitGruppen
	 * 
	 * @param zgs
	 *          Liste der Zeitgruppen
	 */
	@SuppressWarnings("unused")
	private void setListeZeitGruppen(
			List<KalenderZeitFilterDatenStrukturZeitGruppe> zgs)
	{
		_listeZeitGruppen = zgs;
	}

	/**
	 * Liefert einen Clone dieser Klasse. Funktioniert intern ber
	 * erzeugeFilterString und erzeugeAusFilterString Methode dieser Klasse.
	 */
	public KalenderZeitFilterDatenStruktur clone()
	{
		KalenderZeitFilterDatenStruktur rueckgabe = null;
		String filterString = this.erzeugeFilterString();
		rueckgabe = new KalenderZeitFilterDatenStruktur(filterString);
		return rueckgabe;
	}

}
