/*
 * This file is part of Rahmenwerk.
 *
 * Rahmenwerk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License  as 
 * published by the Free Software Foundation, either version 3 of the
 * License.
 *
 * Rahmenwerk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Rahmenwerk.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 *
 * Copyright (c) 2008 Dambach Werke GmbH.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 * Dambach Werke GmbH - initial API and implementation
 *
 *
 *
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */
package de.bsvrz.buv.rw.basislib.listendarstellung.ktableerweiterungen;

import java.util.HashMap;

/**
 * Speichert den Gefaltet-Status von Zellen
 * 
 * @author HofmannJ
 * 
 */
public class FaltStatusSpalten
{
	/**
	 * Erweitert KTable um die Fhigkeit, die Faltung von Zellen zu speichern.
	 * Nimmt den gefaltet Status von Zellen auf. <code>true</code> bedeutet
	 * gefaltet, <code>false</code> bedeutet nicht gefaltet. Der Key-String
	 * beschreibt die Spalte und hat das Muster x;y, wobei x = Nummer der Spalte
	 * und y = Nummer der Zeile
	 */
	private HashMap<String, SpaltenStatus> _map = new HashMap<String, SpaltenStatus>();

	/**
	 * Konstruktor
	 */
	public FaltStatusSpalten()
	{
	}

	/**
	 * Setzt den Status einer Zelle auf gefaltet oder nicht gefaltet.
	 * 
	 * @param x
	 *          X-Position (Spalte) der Zelle
	 * @param Y
	 *          Y-Position (Zeile) der Zelle
	 * @param ausgangsBreite
	 *          Ursprngliche Breite der Zelle
	 * @param gefaltet
	 *          Wenn <code>true</code>, dann ist die Zelle gefaltet markiert,
	 *          bei <code>false</code> ist die Zelle als nicht-gefaltet
	 *          markiert.
	 */
	public void setZelleGefaltet(int x, int y, int ausgangsBreite,
			boolean gefaltet)
	{
		String key = x + ";" + y;
		if (!_map.containsKey(key))
		{
			SpaltenStatus spaltenStatus = new SpaltenStatus(ausgangsBreite, gefaltet);
			_map.put(key, spaltenStatus);
		} else
		{
			SpaltenStatus spaltenStatus = _map.get(key);
			spaltenStatus.setGefaltet(gefaltet);
		}
	}

	/**
	 * Prft, ob eine Zelle gefaltet ist oder nicht. Bisher nicht gesetzte Zellen
	 * werden als nicht-gefaltet gefhrt.
	 * 
	 * @param x
	 *          X-Position (Spalte) der Zelle *
	 * @param Y
	 *          Y-Position (Zeile) der Zelle
	 * @return Liefert <code>true</code>, wenn die Zelle als gefaltet markiert
	 *         ist, sonst <code>false</code>
	 */
	public boolean isZelleGefaltet(int x, int y)
	{
		boolean re = false;
		String key = x + ";" + y;
		if (_map.containsKey(key))
		{
			SpaltenStatus spaltenStatus = _map.get(key);

			re = spaltenStatus.isGefaltet();
		}
		return re;
	}

	/**
	 * Liefert die Ausgangsbreite einer Spalte, sofern diese schon einmal zu
	 * dieser Klasse hinzugefgt wurde.
	 * 
	 * @param x
	 *          Index der Spalte *
	 * @param Y
	 *          Y-Position (Zeile) der Zelle
	 * @return Breite der Spalte in Pixeln oder <code>-1</code>, falls fr die
	 *         angegebene Spalte noch keine Ausgangsbreite angegeben wurde.
	 */
	public int getAusgangsBreite(int x, int y)
	{
		int re = -1;
		String key = x + ";" + y;
		if (_map.containsKey(key))
		{
			re = _map.get(key).getAusgangsBreite();
		}
		return re;
	}

	/**
	 * Hilfsklasse zur Reprsentation von
	 * <ul>
	 * <li>Ausgangsbreite,</li>
	 * <li>Faltstatus (gefaltet oder nicht gefaltet)</li>
	 * </ul>
	 * der jeweiligen Tabellenspalte.
	 * 
	 * @author HofmannJ
	 * 
	 */
	private class SpaltenStatus
	{
		/**
		 * Ausgangsbreite der Spalte
		 */
		private int _ausgangsBreite = 0;

		/**
		 * Flag zur Anzeige, ob die Spalte gefaltet ist oder nicht. Dabei ist
		 * <code>true</code> gefaltet und <code>false</code> nicht gefaltet.
		 */
		private boolean _gefaltet = false;

		/**
		 * Konstruktor, setzt Ausgangsbreite unf Faltstatus
		 * 
		 * @param ausgangsBreite
		 *          Ausgangsbreite der Spalte
		 * @param gefaltet
		 *          Faltungsstatus der Spalte
		 */
		public SpaltenStatus(int ausgangsBreite, boolean gefaltet)
		{
			_ausgangsBreite = ausgangsBreite;
			_gefaltet = gefaltet;
		}

		/**
		 * Liefert die Ursprngliche Breite der Spalte.
		 * 
		 * @return Ursprngliche Breite der Spalte
		 */
		public int getAusgangsBreite()
		{
			return _ausgangsBreite;
		}

		/**
		 * Setzt die ursprngliche Breite der Spalte.
		 * 
		 * @param ausgangsBreite
		 *          Ursprngliche Breite der Spalte
		 */
		@SuppressWarnings("unused")
		public void setAusgangsBreite(int ausgangsBreite)
		{
			this._ausgangsBreite = ausgangsBreite;
		}

		/**
		 * Prft, ob die Spalte gefaltet ist oder nicht.
		 * 
		 * @return Liefert <code>true</code>, wenn die Spalte gefaltet ist, sonst
		 *         <code>false</code>
		 */
		public boolean isGefaltet()
		{
			return _gefaltet;
		}

		/**
		 * Setzt den Faltstatus der jeweiligen Spalte.
		 * 
		 * @param gefaltet
		 *          Wenn <code>true</code>, dann ist die Spalte gefaltet, bei
		 *          <code>false</code> ist die Spalte nicht gefaltet.
		 */
		public void setGefaltet(boolean gefaltet)
		{
			this._gefaltet = gefaltet;
		}

	}

}
