/*
 * This file is part of Rahmenwerk.
 *
 * Rahmenwerk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License  as 
 * published by the Free Software Foundation, either version 3 of the
 * License.
 *
 * Rahmenwerk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Rahmenwerk.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 *
 * Copyright (c) 2008 Dambach Werke GmbH.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 * Dambach Werke GmbH - initial API and implementation
 *
 *
 *
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */

package de.bsvrz.buv.rw.basislib.ort.hilfsklassen;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.Vector;

import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.Data.Array;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.debug.Debug;

public class StrassenTeilSegment
{

	/**
	 * Der Debugger
	 */
	private Debug _debug = Debug.getLogger();

	/**
	 * Das zum Straenteilsegment gehrende Systemobjekt
	 */
	private SystemObject m_systemObject;

	/**
	 * Lnge des Straenteilsegments
	 */
	private long m_laenge = 0;

	/**
	 * Anzahl Fahrstreifen
	 */
	private int m_anzahlFahrstreifen = 0;

	/**
	 * Steigung/Geflle
	 */
	private String m_steigungGefaelle = null;

	/**
	 * Offset des Straenteilsegmentes Innerhalb des Straensegments
	 */
	private long m_offset = 0;

	/**
	 * Datenmodell des DaV
	 */
	private DataModel m_dataModel;

	/**
	 * Liste der Betriebskilometer dieses StraenTeilSegments.
	 */
	private LinkedList<BetriebsKilometerEintrag> m_betriebsKilometer = new LinkedList<BetriebsKilometerEintrag>();

	/**
	 * Liste der ASB-Stationierungen
	 */
	private LinkedList<AsbStationierung> m_asbStationierungen = new LinkedList<AsbStationierung>();

	/**
	 * Referenz auf das zugehrige StraenSegment
	 */
	private StrassenSegment m_strassenSegment = null;

	/**
	 * Konstruktor der Klasse
	 * 
	 * @param dataModel
	 *          Datenmodell des DaV
	 * @param object
	 *          Systemobjekt vom Typ 'typ.straenTeilSegment'
	 * @param strassenSegment
	 *          Referenz auf das StraenSegment, zu dem dieses StraenTeilsegment
	 *          gehrt
	 */
	public StrassenTeilSegment(DataModel dataModel, SystemObject object,
			StrassenSegment strassenSegment)
	{
		m_systemObject = object;

		m_dataModel = dataModel;

		this.m_strassenSegment = strassenSegment;

		AttributeGroup atg = m_dataModel
				.getAttributeGroup("atg.straenTeilSegment");

		if (atg != null)
		{
			Data confData = object.getConfigurationData(atg);

			if (confData != null)
			{
				m_laenge = confData.getUnscaledValue("Lnge").longValue() / 100;
				m_anzahlFahrstreifen = confData.getUnscaledValue("AnzahlFahrStreifen")
						.intValue();
				m_steigungGefaelle = confData.getScaledValue("SteigungGeflle")
						.getValueText();
			}
		}

		/*
		 * Infos zu den Betriebskilometern herausziehen
		 */
		AttributeGroup atgBK = m_dataModel
				.getAttributeGroup("atg.betriebsKilometer");
		if (atgBK != null)
		{
			Data data = object.getConfigurationData(atgBK);
			if (data != null)
			{
				Data itemBetriebsKilometer = data.getItem("BetriebsKilometer");
				if (itemBetriebsKilometer != null)
				{
					Array arrayBetriebsKilometer = data.getArray("BetriebsKilometer");

					for (int i = 0; i < arrayBetriebsKilometer.getLength(); i++)
					{
						Data item = arrayBetriebsKilometer.getItem(i);
						long bkOffset = item.getUnscaledValue("Offset").longValue();
						long bkWert = item.getUnscaledValue("Wert").longValue();

						String blockNummer = item.getTextValue("BlockNummer").getText();

						BetriebsKilometerEintrag bk = new BetriebsKilometerEintrag(
								bkOffset, bkWert, blockNummer);
						m_betriebsKilometer.add(bk);
					}

				}
			}
		}

		/*
		 * Infos zu den ASB-Stationierungen herausziehen
		 */
		AttributeGroup atgAsb = m_dataModel
				.getAttributeGroup("atg.asbStationierung");
		if (atgAsb != null)
		{
			Data data = object.getConfigurationData(atgAsb);
			if (data != null)
			{
				Data itemAsbStationierung = data.getItem("AsbStationierung");
				if (itemAsbStationierung != null)
				{
					Array arrayAsbStationierung = data.getArray("AsbStationierung");
					for (int i = 0; i < arrayAsbStationierung.getLength(); i++)
					{
						Data item = arrayAsbStationierung.getItem(i);
						long offset = item.getUnscaledValue("Offset").longValue();
						String anfangsKnoten = item.getTextValue("AnfangsKnoten").getText();
						String endKnoten = item.getTextValue("EndKnoten").getText();
						long anfang = item.getUnscaledValue("Anfang").longValue();
						long ende = item.getUnscaledValue("Ende").longValue();

						int verkehrsRichtung = item.getUnscaledValue("VerkehrsRichtung")
								.intValue();

						EnumAsbVerkehrsRichtung richtung = null;
						switch (verkehrsRichtung)
						{
						case 0:
							richtung = EnumAsbVerkehrsRichtung.UNBEKANNT;
							break;
						case 1:
							richtung = EnumAsbVerkehrsRichtung.INSTATIONIERUNGSRICHTUNG;
							break;
						case 2:
							richtung = EnumAsbVerkehrsRichtung.GEGENSTATIONIERUNGSRICHTUNG;
							break;
						default:
							richtung = EnumAsbVerkehrsRichtung.UNBEKANNT;
							break;
						}

						AsbStationierung asbStationierung = new AsbStationierung(offset,
								anfangsKnoten, endKnoten, anfang, ende, richtung, this);
						m_asbStationierungen.add(asbStationierung);
					}
				}
			}

		}
	}

	/**
	 * @return liefert die Variable systemObject zurck
	 */
	public SystemObject getSystemObject()
	{
		return m_systemObject;
	}

	/**
	 * Methode liefert die Pid des Straenteilsegments
	 * 
	 * @return Pid des Straenteilsegments
	 */
	public String getPid()
	{
		if (m_systemObject != null)
			return m_systemObject.getPid();

		return null;
	}

	/**
	 * @return liefert die Variable anzahlFahrstreifen zurck
	 */
	public int getAnzahlFahrstreifen()
	{
		return m_anzahlFahrstreifen;
	}

	/**
	 * @return liefert die Variable lnge zurck
	 */
	public long getLaenge()
	{
		return m_laenge;
	}

	/**
	 * @return liefert die Variable steigungGeflle zurck
	 */
	public String getSteigungGefaelle()
	{
		return m_steigungGefaelle;
	}

	/**
	 * Liefert den Offset des Straenteilsegmentes innerhalb des Straensegmentes
	 * zurck.
	 * 
	 * @return liefert die Variable offset zurck
	 */
	public long getOffset()
	{
		return m_offset;
	}

	/**
	 * Setzt den Offset des Straenteilsegmentes innerhalb des Straensegmentes.
	 * 
	 * @param offset
	 *          setzt die Variable offset
	 */
	public void setOffset(long offset)
	{
		m_offset = offset;
	}

	// /**
	// * Liefert den Parameter <code>Wert</code> der
	// * <code>atg.betriebsKilometer</code> (in Metern).
	// *
	// * @return Parameter <code>Wert</code> der
	// * <code>atg.betriebsKilometer</code> (in Metern)
	// */
	// public long getBKWert()
	// {
	// return m_bkWert;
	// }
	//
	// /**
	// * Liefert den Parameter <code>BlockNummer</code> der
	// * <code>atg.betriebsKilometer</code> als <code>long</code>. <br>
	// * Normalerweise wird vom DaV hierzu nur ein String zurckgegben, der Wert
	// * <b>-1</b> ist gleichbedeutend mit <b>KeineAngabe</b>.
	// *
	// * @return Parameter <code>BlockNummer</code> der
	// * <code>atg.betriebsKilometer</code>
	// */
	// public long getBKBlockNummer()
	// {
	// return m_bkBlockNummer;
	// }
	//	
	/**
	 * Liefert die Liste der Betriebskilometer dieses StraenTeilSegmentes zurck.
	 * 
	 * @return Liste der Betriebskilometer.
	 */
	public LinkedList<BetriebsKilometerEintrag> getBetriebsKilometer()
	{
		return m_betriebsKilometer;
	}

	public LinkedList<AsbStationierung> getAsbStationierungen()
	{
		return m_asbStationierungen;
	}

	/**
	 * Fgt einen {@link BetriebsKilometerEintrag} zur Liste der
	 * BetriebsKilometerEintrge hinzu.
	 * 
	 * @param bke
	 *          Hinzuzufgender Eintrag
	 */
	public void hinzufuegenBetriebsKilometerEintrag(BetriebsKilometerEintrag bke)
	{
		if (!m_betriebsKilometer.contains(bke))
		{
			m_betriebsKilometer.add(bke);
		}
	}

	public String toString()
	{
		String re = "";

		// re = "PID: " + m_systemObject.getPid() + " Offset: " + m_bkOffset
		// + " m Wert: " + m_bkWert + " m BlockNummer: " + m_bkBlockNummer;

		re = re + "StraenTeilSegment (Pid: " + m_systemObject.getPid()
				+ ") Lnge = " + m_laenge + " m :\n";

		Iterator<BetriebsKilometerEintrag> itBK = m_betriebsKilometer.iterator();
		while (itBK.hasNext())
		{
			re = re + "\t=> " + itBK.next().toString() + "\n";
		}

		return re;
	}

	/**
	 * Bestimmt die Bereiche dieses StraenTeilsegmentes
	 */
	public String bestimmeBereiche()
	{
		StringBuffer sb = new StringBuffer();

		Vector<StrassenTeilSegment> strassenTeilSegmente = m_strassenSegment
				.getStrassenTeilSegmente();

		Iterator<BetriebsKilometerEintrag> it = m_betriebsKilometer.iterator();
		while (it.hasNext())
		{
			BetriebsKilometerEintrag bk = it.next();

			sb.append("BKE: ");
			sb.append(bk.toString());
			sb.append("\n");

			sb.append("StrassenSegment: ");
			sb.append(m_strassenSegment.getPid());
			sb.append(" Lnge: ");
			sb.append(m_strassenSegment.getLaenge());
			sb.append("\n");

			if (it.hasNext())
			{
				// Sonderfall, mehrere BeteriebKilometerEintrge in einem STS
			} else
			{
				// Nachsehen in den nachfolgenden StraenTeilSegmenten dieses
				// StraenSegments
				if (strassenTeilSegmente.contains(this))
				{
					/*
					 * Index dieses STS innerhalb der Liste der StraenTeilSegmente
					 */
					int idxThis = strassenTeilSegmente.indexOf(this);
					int laengeStrassenTeilSegmente = strassenTeilSegmente.size();

					sb.append("IDX THIS: ");
					sb.append(idxThis);
					sb.append(" SIZE: ");
					sb.append(laengeStrassenTeilSegmente);
					sb.append("\n");

					if (idxThis <= laengeStrassenTeilSegmente - 2)
					{
						sb
								.append("\tBestimmen innerhalb der Betriebskilometer des gleichen STS\n");

						StrassenTeilSegment strassenTeilSegment = strassenTeilSegmente
								.get(idxThis + 1);

						if (strassenTeilSegment.getBetriebsKilometer().size() > 0)
						{
							BetriebsKilometerEintrag bkNext = strassenTeilSegment
									.getBetriebsKilometer().get(0);

							sb.append(bkNext.toString());
							sb.append("\n");
						}

					} else
					{
						/*
						 * Ansonsten: im ersten Straensegment des nchsten STS den ersten
						 * BK bestimmen
						 */
						long bestimmeVomNaechstenStrassenSegmentDenErstenStSBkWert = bestimmeVomNaechstenStrassenSegmentDenErstenStSBkWert();
						sb.append("VomNaechstenStrassenSegmentDenErstenStSBkWert:");
						sb.append(bestimmeVomNaechstenStrassenSegmentDenErstenStSBkWert);
						sb.append("\n");

					}
				}

			}
			// else
			// {
			// System.out.println("HAT");
			// }
		}

		return sb.toString();
	}

	/**
	 * Bestimmt vom Nachfolger des StraenSegmentes (in dem sich dieses
	 * StraenTeilSegment befindet, ber einen StraenKnoten hinweg) vom ersten
	 * StaenTeilSegment den Wert des ersten Betriebskilometereintrag.
	 * 
	 * @return Wert des ersten Betriebskilometereintrags des nachfolgenden
	 *         Straensegments oder <code>-1</code>, falls dieser nicht ermittelt
	 *         werden konnte.
	 */
	private long bestimmeVomNaechstenStrassenSegmentDenErstenStSBkWert()
	{
		long re = -1;

		_debug.fine("Bestimmen innerhalb des ersten BKs des nchsten STS");
		AeusseresStrassenSegment aeusseresStrassenSegment = (AeusseresStrassenSegment) m_strassenSegment;
		int groesseNachfolgerAufStrasse = aeusseresStrassenSegment
				.getNachfolgerAufStrasse().size();
		_debug.fine("ANZAHL NACHFOLGER: " + groesseNachfolgerAufStrasse);
		_debug.fine("RICHTUNG         : "
				+ aeusseresStrassenSegment.getTmcRichtung());

		// if (groesseNachfolgerAufStrasse == 1)
		// {
		//			
		// } else if (groesseNachfolgerAufStrasse > 1)
		// {
		Iterator<StrassenSegment> it = aeusseresStrassenSegment
				.getNachfolgerAufStrasse().iterator();
		while (it.hasNext())
		{
			StrassenSegment strassenSegment = it.next();
			_debug.fine("===>InneresStraenSegment in einem Knoten");
			_debug.fine("===>TYP:" + strassenSegment.getClass().getName());
			if (strassenSegment instanceof AeusseresStrassenSegment)
			{
				AeusseresStrassenSegment intern = (AeusseresStrassenSegment) strassenSegment;
				_debug.fine("===>A: " + intern.getPid() + " RICHTUNG: "
						+ intern.getRichtungStrasse());
			} else if (strassenSegment instanceof InneresStrassenSegment)
			{
				InneresStrassenSegment inneresStrassenSegment = (InneresStrassenSegment) strassenSegment;
				_debug.fine("===>InnersesStraenSegment: "
						+ inneresStrassenSegment.getPidRichtungStrasse()
						+ "Nachfolger auf Strae, Anzahl Nachfolger: "
						+ inneresStrassenSegment.getNachfolgerAufStrasse().size());
				Iterator<StrassenSegment> itNachfolgerInneresseStrassenSegment = inneresStrassenSegment
						.getNachfolgerAufStrasse().iterator();
				while (itNachfolgerInneresseStrassenSegment.hasNext())
				{
					StrassenSegment ssnachfolger = itNachfolgerInneresseStrassenSegment
							.next();
					if (ssnachfolger instanceof AeusseresStrassenSegment)
					{
						_debug.fine("====>Nachfolgendes ueres StraenSegment");
						AeusseresStrassenSegment ausseresStrassenSegmentsNachInnerenSegment = (AeusseresStrassenSegment) ssnachfolger;
						_debug.fine("====>AUSSEN: "
								+ ausseresStrassenSegmentsNachInnerenSegment.getPid());
						_debug.fine("====>RICHTUNG: "
								+ ausseresStrassenSegmentsNachInnerenSegment.getTmcRichtung());

						Vector<StrassenTeilSegment> strassenTeilSegmente = aeusseresStrassenSegment
								.getStrassenTeilSegmente();
						if ((strassenTeilSegmente != null)
								&& (strassenTeilSegmente.size() > 0))
						{
							StrassenTeilSegment strassenTeilSegment = strassenTeilSegmente
									.get(0);
							LinkedList<BetriebsKilometerEintrag> betriebsKilometer = strassenTeilSegment
									.getBetriebsKilometer();
							if ((betriebsKilometer != null) && (betriebsKilometer.size() > 0))
							{
								long wert = betriebsKilometer.get(0).getWert();

								if (re == -1)
								{
									re = wert;
								} else if (re == wert)
								{
									// alles ok
								} else if (re != wert)
								{
									_debug.fine("Problem mit Paralellfahrbahn");
								}

								_debug.fine("========>STS: " + strassenTeilSegment.getPid()
										+ "Lnge:" + strassenTeilSegment.getLaenge() + "<===");
								_debug.fine("========>WERT: " + wert + " <========");
							}
						}

					}

				}

			}
		}
		// }

		return re;
	}

	/**
	 * Diese Klasse reprsentiert einen Eintrag in der ArrayLiest der
	 * atg.betriebsKilometer. <br>
	 * Mit Hilfe des Konstruktors wird der Eintrag angelegt, die entsprechenden
	 * Getter-Methoden liefern die Werte. Setter-Methoden sind nicht vorgesehen.
	 * 
	 * @author HofmannJ
	 * 
	 */
	public class BetriebsKilometerEintrag
	{
		private long _offset = 0;

		private long _wert = -1;

		private String _blockNummer = "KeineAngabe";

		/**
		 * Gibt an, ob dieser Betriebskilometereintrag berechnet wurde.
		 */
		private boolean _berechnet = false;

		/**
		 * Konstruktor fr den {@link BetriebsKilometerEintrag}. bernimmt Offset,
		 * Wert und BlockNummer als Parameter.
		 * 
		 * @param offset
		 *          Offset Offset Offset, ab dem der Betriebskilometereintrag in
		 *          diesem StraenTeilSegment gltig ist.
		 * @param wert
		 *          Wert des Betriebskilometers (in Metern)
		 * @param blockNummer
		 *          Blocknummer Blocknummer des fr dieses StraenTeilSegments
		 *          gltigen Betriebskilometers.
		 */
		public BetriebsKilometerEintrag(long offset, long wert, String blockNummer)
		{
			this._offset = offset;
			this._wert = wert;
			this._blockNummer = blockNummer;
		}

		/**
		 * Konstruktor fr den {@link BetriebsKilometerEintrag}. bernimmt Offset,
		 * Wert und BlockNummer als Parameter.
		 * 
		 * @param offset
		 *          Offset Offset, ab dem der Betriebskilometereintrag in diesem
		 *          StraenTeilSegment gltig ist.
		 * @param wert
		 *          Wert des Betriebskilometers (in Metern)
		 * @param blockNummer
		 *          Blocknummer Blocknummer des fr dieses StraenTeilSegments
		 *          gltigen Betriebskilometers.
		 * @param berechnet
		 *          Gibt an, ob dieser Betriebskilometereintrag berechnet wurde oder
		 *          nicht. Muss <code>true</code> sein, wenn er berechnet wurde,
		 *          sonst <code>false</code>.
		 */
		public BetriebsKilometerEintrag(long offset, long wert, String blockNummer,
				boolean berechnet)
		{
			this._offset = offset;
			this._wert = wert;
			this._blockNummer = blockNummer;
			this._berechnet = berechnet;
		}

		/**
		 * Liefert den Offset zurck.
		 * 
		 * @return Offset.
		 */
		public long getOffset()
		{
			return _offset;
		}

		/**
		 * Liefert den Wert zurck. Dieser wird in Metern geliefert.
		 * 
		 * @return Wert in Metern.
		 */
		public long getWert()
		{
			return _wert;
		}

		/**
		 * Liefert die BlockNummer zurck.
		 * 
		 * @return Liefert die BlockNummer oder -1, wenn kein Eintrag vorhanden ist.
		 */
		public String getBlockNummer()
		{
			return _blockNummer;
		}

		/**
		 * Erfragt, ob dieser {@link BetriebsKilometerEintrag} berechnet wurde oder
		 * nicht.
		 * 
		 * @return Liefert <code>true</code>, wenn dieser
		 *         {@link BetriebsKilometerEintrag} berechnet wurde, sonst
		 *         <code>false</code>.
		 */
		public boolean wurdeBerechnet()
		{
			return _berechnet;
		}

		public String toString()
		{
			String re = "";
			re = re + "Offset: " + _offset + " m Wert: " + _wert + " m BlockNummer: "
					+ _blockNummer;
			return re;
		}
	}

	/**
	 * Speichert die Daten der atg.asbStationierung.
	 * 
	 * @author HofmannJ
	 * 
	 */
	public class AsbStationierung
	{
		/**
		 * Abstand des ASB-Stationierungs-Punktes vom Anfang des Linienobjekts.
		 */
		private long _offset = -1;

		/**
		 * Anfangsknoten mit eindeutiger Kennung (Anfangsnullpunkt) des
		 * Teilabschnittes oder Astes.
		 */
		private String _anfangsKnoten = null;

		/**
		 * Endknoten mit eindeutiger Kennung (Endnullpunkt) des Teilabschnittes oder
		 * Astes.
		 */
		private String _endKnoten = null;

		/**
		 * Anfangsstationierung. Stationierungsangabe in Metern relativ zum
		 * Abschnitt oder Ast.
		 */
		private long _anfang = -1;

		/**
		 * Endstationierung. Stationierungsangabe in Metern relativ zum Abschnitt
		 * oder Ast.
		 */
		private long _ende = -1;

		/**
		 * Referenz auf das StrassenTeilSegment, zu dem diese ASB-Stationierung
		 * gehrt.
		 */
		private StrassenTeilSegment _gehoertZuStrassenTeilSegment = null;

		/**
		 * Angabe des Richtungsbezugs der ASB-Stationierung relativ zur Richtung des
		 * Straensegments.
		 */
		private EnumAsbVerkehrsRichtung _verkehrsRichtung = null;

		/**
		 * Konstruktor. bernimmt die nachfolgend aufgelisteten Parameter zur
		 * Initialisierung eines {@link AsbStationierung}-Objektes:
		 * 
		 * @param offset
		 *          Abstand des ASB-Stationierungs-Punktes vom Anfang des
		 *          Linienobjekts.
		 * @param anfangsKnoten
		 *          Anfangsknoten mit eindeutiger Kennung (Anfangsnullpunkt) des
		 *          Teilabschnittes oder Astes. Die Nullpunktbezeichung hat die Form
		 *          TTTTnnnB, wobei TTTT die vierstellige TK25-Blattnummer und nnn
		 *          die dreistellige laufende Nummer ist, die zusammen die
		 *          bundesweit eindeutige Netzknotennummer darstellen. Durch die
		 *          Kennung B (ein Zeichen) wird zustzlich der Nullpunkt des
		 *          Abschnitts oder Astes eindeutig festgelegt. Eine nicht
		 *          vorhandene Kennung wird als o(hne) eingetragen.
		 * @param endKnoten
		 *          Endknoten mit eindeutiger Kennung (Endnullpunkt) des
		 *          Teilabschnittes oder Astes. Die Nullpunktbezeichung hat die Form
		 *          TTTTnnnB, wobei TTTT die vierstellige TK25-Blattnummer und nnn
		 *          die dreistellige laufende Nummer ist, die zusammen die
		 *          bundesweit eindeutige Netzknotennummer darstellen. Durch die
		 *          Kennung B (ein Zeichen) wird zustzlich der Nullpunkt des
		 *          Abschnitts oder Astes eindeutig festgelegt. Eine nicht
		 *          vorhandene Kennung wird als o(hne) eingetragen.
		 * @param anfang
		 *          Anfangsstationierung. Stationierungsangabe in Metern relativ zum
		 *          Abschnitt oder Ast.
		 * @param ende
		 *          Endstationierung. Stationierungsangabe in Metern relativ zum
		 *          Abschnitt oder Ast.
		 * @param verkehrsRichtung
		 *          Angabe des Richtungsbezugs der ASB-Stationierung relativ zur
		 *          Richtung des Straensegments.
		 * @param gehoertZuStrassenTeilSegment
		 *          Referenz auf das {@link StrassenTeilSegment}, zus dem diese
		 *          {@link AsbStationierung} gehrt.s
		 */
		public AsbStationierung(long offset, String anfangsKnoten,
				String endKnoten, long anfang, long ende,
				EnumAsbVerkehrsRichtung verkehrsRichtung,
				StrassenTeilSegment gehoertZuStrassenTeilSegment)
		{
			this._offset = offset;
			this._anfangsKnoten = anfangsKnoten;
			this._endKnoten = endKnoten;
			this._anfang = anfang;
			this._ende = ende;
			this._verkehrsRichtung = verkehrsRichtung;
			this._gehoertZuStrassenTeilSegment = gehoertZuStrassenTeilSegment;
		}

		/**
		 * Liefert den Abstand des ASB-Stationierungs-Punktes vom Anfang des
		 * Linienobjekts.
		 * 
		 * @return Abstand des ASB-Stationierungs-Punktes vom Anfang des
		 *         Linienobjekts
		 */
		public long getOffset()
		{
			return _offset;
		}

		/**
		 * Setzt den Abstand des ASB-Stationierungs-Punktes vom Anfang des
		 * Linienobjekts.
		 * 
		 * @param offset
		 *          Abstand des ASB-Stationierungs-Punktes vom Anfang des
		 *          Linienobjekts
		 */
		public void setOffset(long offset)
		{
			this._offset = offset;
		}

		/**
		 * Liefert den String, der den Anfangsknoten definiert.
		 * 
		 * @return Anfangsknotendefinition.
		 */
		public String getAnfangsKnoten()
		{
			return _anfangsKnoten;
		}

		/**
		 * Setzt den String, der den Anfangsknoten definiert.
		 * 
		 * @param anfangsKnoten
		 *          Anfangsknotendefinition.
		 */
		public void setAnfangsKnoten(String anfangsKnoten)
		{
			this._anfangsKnoten = anfangsKnoten;
		}

		/**
		 * Liefert vom Anfangsknoten die (4-stellige) <b>TK25-Blattnummer</b>.
		 * 
		 * @return TK25-Blattnummer oder <code>-1</code>, falls diese nicht
		 *         ermittelt werden konnte.
		 */
		public int getAnfangsKnotenBlattNr()
		{
			int re = -1;

			if (_anfangsKnoten != null)
			{
				String sub = _anfangsKnoten.substring(0, 4);
				re = Integer.valueOf(sub);
			}

			return re;
		}

		/**
		 * Liefert die laufende Nummer vom Anfangsknoten.
		 * 
		 * @return Laufende Nummer vom Anfangsknoten oder <code>-1</code>, falls
		 *         diese nicht ermittelt werden konnte.
		 */
		public int getAnfangsKnotenLaufendeNummer()
		{
			int re = -1;

			if (_anfangsKnoten != null)
			{
				String sub = _anfangsKnoten.substring(4, 7);
				re = Integer.valueOf(sub);
			}

			return re;
		}

		/**
		 * Prft, ob es sich bei dem Angangsknoten um einen Nullpunkt handelt.
		 * 
		 * @return Liefert <code>true</code>, wenn es sich bei dem Angangsknoten um
		 *         einen Nullpunkt handelt, sonst <code>false</code>.<br>
		 *         Wenn kein Anfangsknoten ermittelt werden konnte, wird
		 *         <code>false</code> geliefert.
		 */
		public boolean isAnfangsKnotenNullpunkt()
		{
			boolean re = false;

			if (_anfangsKnoten != null)
			{
				try
				{
					char c = _anfangsKnoten.charAt(7);
					if (c == 'B')
					{
						re = true;
					} else
					{
						re = false;
					}
				} catch (IndexOutOfBoundsException e)
				{
					_debug.fine(getClass().getName()
							+ "::isAnfangsKnotenNullpunkt \n\t=> " + e.toString());
				}

			}

			return re;
		}

		/**
		 * Liefert den String, der den Endknoten definiert.
		 * 
		 * @return Endknotendefinition.
		 */
		public String getEndKnoten()
		{
			return _endKnoten;
		}

		/**
		 * Setzt den String, der den Endknoten definiert.
		 * 
		 * @param endKnoten
		 *          Endknotendefinition
		 */
		public void setEndKnoten(String endKnoten)
		{
			this._endKnoten = endKnoten;
		}

		/**
		 * Liefert vom Endknoten die (4-stellige) <b>TK25-Blattnummer</b>.
		 * 
		 * @return TK25-Blattnummer oder <code>-1</code>, falls diese nicht
		 *         ermittelt werden konnte.
		 */
		public int getEndKnotenBlattNr()
		{
			int re = -1;

			if (_endKnoten != null)
			{
				String sub = _endKnoten.substring(0, 4);
				re = Integer.valueOf(sub);
			}

			return re;
		}

		/**
		 * Liefert die laufende Nummer vom Endknoten.
		 * 
		 * @return Laufende Nummer vom Endsknoten oder <code>-1</code>, falls diese
		 *         nicht ermittelt werden konnte.
		 */
		public int getEndKnotenLaufendeNummer()
		{
			int re = -1;

			if (_endKnoten != null)
			{
				String sub = _endKnoten.substring(4, 7);
				re = Integer.valueOf(sub);
			}

			return re;
		}

		/**
		 * Liefert das {@link StrassenTeilSegment}, zu dem diese
		 * {@link AsbStationierung} gehrt.
		 * 
		 * @return {@link StrassenTeilSegment}, zu dem diese
		 *         {@link AsbStationierung} gehrt
		 */
		public StrassenTeilSegment getGehoertZuStrassenTeilSegement()
		{
			return _gehoertZuStrassenTeilSegment;
		}

		/**
		 * Prft, ob es sich bei dem Endknoten um einen Nullpunkt handelt.
		 * 
		 * @return Liefert <code>true</code>, wenn es sich bei dem Endknoten um
		 *         einen Nullpunkt handelt, sonst <code>false</code>.<br>
		 *         Wenn kein Endknoten ermittelt werden konnte, wird
		 *         <code>false</code> geliefert.
		 */
		public boolean isEndKnotenNullpunkt()
		{
			boolean re = false;

			if (_endKnoten != null)
			{
				try
				{
					char c = _endKnoten.charAt(7);
					if (c == 'B')
					{
						re = true;
					} else
					{
						re = false;
					}
				} catch (IndexOutOfBoundsException e)
				{
					_debug.error(getClass().getName() + "::isEndKnotenNullpunkt \n\t=> "
							+ e.toString());
				}
			}

			return re;
		}

		/**
		 * Liefert die Anfangsstationierung (Stationierungsangabe in Metern relativ
		 * zum Abschnitt oder Ast).
		 * 
		 * @return Anfangsstationierung (Stationierungsangabe in Metern relativ zum
		 *         Abschnitt oder Ast)
		 */
		public long getAnfang()
		{
			return _anfang;
		}

		/**
		 * Setzt die Anfangsstationierung (Stationierungsangabe in Metern relativ
		 * zum Abschnitt oder Ast).
		 * 
		 * @param anfang
		 *          Anfangsstationierung (Stationierungsangabe in Metern relativ zum
		 *          Abschnitt oder Ast)
		 */
		public void setAnfang(long anfang)
		{
			this._anfang = anfang;
		}

		/**
		 * Liefert die Endstationierung (Stationierungsangabe in Metern relativ zum
		 * Abschnitt oder Ast).
		 * 
		 * @return Endstationierung (Stationierungsangabe in Metern relativ zum
		 *         Abschnitt oder Ast)
		 */
		public long getEnde()
		{
			return _ende;
		}

		/**
		 * Setzt die Endstationierung (Stationierungsangabe in Metern relativ zum
		 * Abschnitt oder Ast).
		 * 
		 * @param ende
		 *          Endstationierung (Stationierungsangabe in Metern relativ zum
		 *          Abschnitt oder Ast)
		 */
		public void setEnde(long ende)
		{
			this._ende = ende;
		}

		/**
		 * Liefert die Verkehrsrichtung als Instanz des Enumerators
		 * {@link EnumAsbVerkehrsRichtung}.
		 * 
		 * @return {@link EnumAsbVerkehrsRichtung}
		 */
		public EnumAsbVerkehrsRichtung getVerkehrsRichtung()
		{
			return _verkehrsRichtung;
		}

		/**
		 * Setzt die Verkehrsrichtung als Instanz des Enumerators
		 * {@link EnumAsbVerkehrsRichtung}.
		 * 
		 * @param verkehrsRichtung
		 *          {@link EnumAsbVerkehrsRichtung}
		 */
		public void setVerkehrsRichtung(EnumAsbVerkehrsRichtung verkehrsRichtung)
		{
			this._verkehrsRichtung = verkehrsRichtung;
		}

		public String toString()
		{
			String re = "";
			re = re + "===" + getClass().getName() + "===\n";
			re = re + "Offset:           " + _offset + "\n";
			re = re + "AnfangsKnoten:    " + _anfangsKnoten + "\n";
			re = re + "EndKnoten:        " + _endKnoten + "\n";
			re = re + "Anfang:           " + _anfang + "\n";
			re = re + "Ende:             " + _ende + "\n";
			re = re + "Verkehrsrichtung: ";
			if (_verkehrsRichtung != null)
			{
				switch (_verkehrsRichtung)
				{
				case UNBEKANNT:
					re = re + "unbekannt";
					break;
				case INSTATIONIERUNGSRICHTUNG:
					re = re + "inStationierungsRichtung";
					break;
				case GEGENSTATIONIERUNGSRICHTUNG:
					re = re + "gegenStationierungsRichtung";
					break;
				default:
					re = re + "unbekannt";
					break;
				}
			} else
			{
				re = re + "null";
			}
			re = re + "\n";

			for (int i = 0; i < (6 + getClass().getName().length()); i++)
			{
				re = re + "=";
			}
			re = re + "\n";

			return re;
		}

	}

}
