/*
 * Copyright 2007 by Kappich Systemberatung, Aachen
 * Copyright 2005 by Kappich+Kni Systemberatung Aachen (K2S)
 * 
 * This file is part of de.bsvrz.dav.daf.
 * 
 * de.bsvrz.dav.daf is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.dav.daf is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with de.bsvrz.dav.daf; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package de.bsvrz.dav.daf.main.archive;

import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.config.SystemObject;

import java.util.Collection;
import java.util.List;

/**
 * Schnittstelle, die von Applikationen benutzt wird, um Anfragen an ein Archivsystem zu stellen. Eine konkrete
 * Implementierung wird von den Datenverteiler-Applikationsfunktionen nach erfolgreichem Verbindungsaufbau zum
 * Datenverteiler ber die Methode {@link de.bsvrz.dav.daf.main.ClientDavInterface#getArchive} zur Verfgung gestellt.
 *
 * @author Kappich Systemberatung
 * @version $Revision: 5064 $
 */
public interface ArchiveRequestManager {

	/**
	 * Spezifiziert die Gre des Empfangspuffers je Archivanfrage, die mit nachfolgenden <code>request</code> Aufrufen
	 * erzeugt werden. Der angegebene Parameter ist eine grobe Vorgabe fr alle Streams einer Archivanfrage zusammen und
	 * muss von einer Implementierung nicht exakt bercksichtigt werden, sondern sollte in der sendeseitigen Blockbildung
	 * und in der Flusskontrolle so bercksichtigt werden, dass die Summe der vom Archivsystem gesendeten aber noch nicht
	 * von der anfragenden Applikation abgerufenen Bytes in der angegebenen Grenordnung liegt. Der Defaultwert ist 0 mit
	 * der Bedeutung, das die Gre durch einen entsprechenden Parameter des Archivsystems festgelegt wird.
	 *
	 * @param numberOfBytes Grobe Vorgabe bezglich der Anzahl zu puffernden Bytes pro Archivanfrage oder 0, wenn keine
	 *                      Vorgabe seitens der Empfangsapplikation gemacht werden soll und statt dessen ein entsprechender
	 *                      Parameter des Archivsystems verwendet werden soll.
	 */
	void setReceiveBufferSize(int numberOfBytes);

	/**
	 * Ruft Archivdaten von einem Archivsystem mit Hilfe eines Ergebnisdatenstroms ab. Diese Methode wird von einer
	 * Applikation aufgerufen um Archivdaten von einem Archivsystem abzufragen. Eine Implementierung dieser Methode sollte
	 * ein Objekt zurckliefern ber das asynchron auf das Ergebnis der Archivanfrage zugegriffen werden kann. Im Ergebnis
	 * wird fr die im Parameter <code>spec</code> spezifizierten Daten ein Stream von Ergebnisdatenstzen erwartet.
	 *
	 * @param priority Prioritt der Anfrage
	 * @param spec     Spezifikation der gewnschten Archivdaten
	 * @return Ergebnisobjekt ber das asynchron auf die gewnschten Ergebnisdatenstze zugegriffen werden kann.
	 * @throws IllegalStateException Das Archivsystem ist nicht erreichbar.
	 */
	ArchiveDataQueryResult request(ArchiveQueryPriority priority, ArchiveDataSpecification spec) throws IllegalStateException;

	/**
	 * Ruft Archivdaten von einem Archivsystem mit Hilfe mehrerer Ergebnisdatenstrme ab. Diese Methode wird von einer
	 * Applikation aufgerufen um Archivdaten von einem Archivsystem abzufragen. Eine Implementierung dieser Methode sollte
	 * ein Objekt zurckliefern ber das asynchron auf die Ergebnisse der Archivanfrage zugegriffen werden kann. Im
	 * Ergebnis wird fr jede Archivdatenspezifikation in der bergebenen Liste <code>specs</code> jeweils ein Stream von
	 * Ergebnisdatenstzen erwartet.
	 *
	 * @param priority Prioritt der Anfrage
	 * @param specs    Liste mit Spezifikationen der gewnschten Archivdaten
	 * @return Ergebnisobjekt ber das asynchron auf die Ergebnisdatenstrme mit den gewnschten Ergebnisdatenstzen
	 *         zugegriffen werden kann.
	 * @throws IllegalStateException Das Archivsystem ist nicht erreichbar.
	 */
	ArchiveDataQueryResult request(ArchiveQueryPriority priority, List<ArchiveDataSpecification> specs) throws IllegalStateException;

	/**
	 * Start einer Archivinformationsanfrage an das Archivsystem. Diese Methode wird von einer Applikation aufgerufen, um
	 * eine Archivinformationsanfrage an das Archivsystem zu starten. Eine Implementierung dieser Methode sollte ohne zu
	 * blockieren ein Objekt zurckliefern ber das asynchron auf das Ergebnis der Anfrage zugegriffen werden kann.
	 * <p/>
	 * Das Ergebnisobjekt beinhaltet eine Liste von Objekten. Diese Objekte geben unter anderem ein Objekt vom Typ
	 * <code>ArchiveDataSpecification</code> zurck. Dieses Objekt ist eine Referenz auf den Eingabeparameter
	 * <code>spec</code>.
	 *
	 * @param spec Spezifikation der Archivdaten zu denen Informationen gewnscht werden.
	 * @return Ergebnisobjekt ber das asynchron auf die gewnschten Informationen zugegriffen werden kann.
	 * @throws IllegalStateException Das Archivsystem ist nicht erreichbar.
	 */
	ArchiveInfoQueryResult requestInfo(ArchiveDataSpecification spec) throws IllegalStateException;

	/**
	 * Die Implementation dieser Methode meldet sich als Empfnger von aktuellen Daten an. Die zurckgegebenen Datenstze
	 * unterscheiden sich von einer "normalen" Anmeldung als Empfnger, da Archivdaten zeitlich vor die aktuellen Daten
	 * gemischt werden. Der Benutzer legt dabei fest wie gross der Zeitraum sein soll, der vor die aktuellen Daten gemischt
	 * werden soll. Der Zeitraum wird als "Zeitspanne" oder als "Anzahl Datenstze" angegeben. </br> Ist kein Archivsystem
	 * vorhanden oder bei der Archivanfrage tritt ein Fehler auf, werden nur die aktuellen Daten zurckgegeben.
	 *
	 * @param receiver        Objekt, das Methoden zur Verfgung stellt um den Strom aus historischen und aktuellen Daten
	 *                        entgegen zu nehmen
	 * @param object          System-Objekt fr die die spezifizierten Daten anzumelden sind
	 * @param dataDescription Beschreibende Informationen zu den anzumeldenden Daten
	 * @param options         Fr die Anmeldung zu verwendende Optionen
	 * @param historyType     <code>HistoryTypeParameter.TIME</code> = Der Parameter <code>history</code> bezieht sich auf
	 *                        einen Zeitraum, der vor den aktuellen Daten liegen soll (in ms);
	 *                        <code>HistoryTypeParameter.INDEX</code> = Der Parameter <code>history</code> bezieht sich auf
	 *                        die Anzahl Datenstze, die mindestens vor den aktuellen Daten liegen sollen
	 * @param history         Zeitraum der Archivdaten in Millisekunden, die vor den ersten aktuellen Datenstzen liegen
	 */
	void subscribeReceiver(DatasetReceiverInterface receiver,
						   SystemObject object,
						   DataDescription dataDescription,
						   ReceiveOptions options,
						   HistoryTypeParameter historyType,
						   long history);

	/**
	 * Die Implementation dieser Methode meldet einen Empfnger ab, der mit der Methode {@link
	 * ArchiveRequestManager#subscribeReceiver} angemeldet wurde.
	 *
	 * @param receiver
	 * @param object          System-Objekt fr die die spezifizierten Daten angemeldet wurden
	 * @param dataDescription Beschreibende Informationen zu den angemeldeten Daten
	 */
	void unsubscribeReceiver(DatasetReceiverInterface receiver,
							 SystemObject object,
							 DataDescription dataDescription);

	/**
	 * Start einer Archivinformationsanfrage an das Archivsystem. Diese Methode wird von einer Applikation aufgerufen, um
	 * eine Archivinformationsanfrage an das Archivsystem zu starten. Eine Implementierung dieser Methode sollte ohne zu
	 * blockieren ein Objekt zurckliefern ber das asynchron auf das Ergebnis der Anfrage zugegriffen werden kann.
	 * <p/>
	 * Das Ergebnisobjekt beinhaltet eine Liste von Objekten. Diese Objekte geben unter anderem ein Objekt vom Typ
	 * <code>ArchiveDataSpecification</code> zurck. Diese Objekte sind Referenzen auf Eintrge des bergabeparameters
	 * <code>specs</code>.
	 *
	 * @param specs Liste mit Spezifikationen der Archivdaten zu denen Informationen gewnscht werden
	 * @return Ergebnisobjekt ber das asynchron auf die gewnschten Informationen zugegriffen werden kann.
	 * @throws IllegalStateException Das Archivsystem ist nicht erreichbar.
	 */
	ArchiveInfoQueryResult requestInfo(List<ArchiveDataSpecification> specs) throws IllegalStateException;

	/**
	 * Beauftragt das Archivsystem Archivdaten der Sicherung {@link de.bsvrz.dav.daf.main.impl.archive.filesaver.ArchiveFileSaver} zu bergeben.
	 * Eine Applikation ruft diese Methode auf, um alle Archivdaten, die gesichert werden knnen, zu sichern. Eine
	 * Implementierung dieser Methode sollte ohne zu blockieren ein Objekt zurckliefern ber das asynchron auf das
	 * Ergebnis der Anfrage zugegriffen werden kann.
	 *
	 * @return Ergebnisobjekt ber das Informationen ber den Zustand des Auftrags abgefragt werden knnen
	 * @throws IllegalStateException Das Archivsystem ist nicht erreichbar.
	 */
	ArchiveQueryResult savePersistentData() throws IllegalStateException;

	/**
	 * Start der Wiederherstellung von Datenstzen, die der Sicherung bergeben wurden (siehe {@link
	 * ArchiveRequestManager#savePersistentData()}). Diese Methode wird von einer Applikation aufgerufen, die Datenstze
	 * bentgt, die sich nicht im direkten Zugriff des Archivsystems befinden, sondern bereits der Sicherung bergeben
	 * wurden und spter gelscht wurden (siehe {@link ArchiveRequestManager#deleteDataSimulationVariant}, {@link
	 * ArchiveRequestManager#deleteData}). Eine Implementierung dieser Methode sollte ohne zu blockieren ein Objekt
	 * zurckliefern ber das asynchron auf das Ergebnis des Auftrags zugegriffen werden kann.
	 *
	 * @param requiredData Eine Liste von Zeitbereichen/Indexbereichen, die Wiederhergestellt werden mssen.
	 * @return Ergebnisobjekt ber das Informationen ber den Zustand des Auftrags abgefragt werden knnen
	 * @throws IllegalStateException Das Archivsystem ist nicht erreichbar.
	 */
	ArchiveQueryResult restorePersistentData(List<ArchiveInformationResult> requiredData) throws IllegalStateException;

	/**
	 * Beauftragt das Archivsystem Archivdaten, die zu einer Simulation gehren, aus dem Archivsystem zu lschen. Der
	 * Lschauftrag bezieht sich dabei nicht auf Datenstze, die bereits der Sicherung bergeben wurden und von dieser
	 * verwaltet werden (siehe {@link ArchiveRequestManager#savePersistentData()}). Die Methode wird von einer Applikation
	 * aufgerufen, um nicht mehr bentigte Datenstze, die zu einer Simulation gehren, aus dem Archivsystem zu lschen.
	 *
	 * @param simulationVariant Simulationsvariante, ganzzahliger Wert zwischen 1,...,999
	 * @return Ergebnisobjekt, ber das Informationen ber den Zustand des Auftrags abgefragt werden knnen
	 * @throws IllegalStateException Das Archivsystem ist nicht erreichbar.
	 */
	ArchiveQueryResult deleteDataSimulationVariant(short simulationVariant) throws IllegalStateException;

	/**
	 * Beauftragt das Archivsystem den Lschzeitpunkt der angegebenen Zeitbereiche zu verlngern. Die Methode wird von
	 * einer Applikation aufgerufen, um bentigte Zeitbereiche lnger als vorgesehen im direkten Zugriff des Archivsystems
	 * zu halten.
	 *
	 * @param requiredData Zeitbereiche, die lnger im direkten Zugriff des Archivsystems bleiben sollen
	 * @param timePeriod   Zeitspanne, die die ausgewhlten Daten lnger im direkten Zugriff des Archivsystem bleiben
	 *                     sollen (in ms)
	 * @return Ergebnisobjekt, ber das Informationen ber den Zustand des Auftrags abgefragt werden knnen
	 */
	ArchiveQueryResult increaseDeleteTime(List<ArchiveInformationResult> requiredData, long timePeriod);

	/**
	 * Beauftragt das Archivsystem die angegebenen Zeitbereiche ohne Rcksicht auf den mit {@link
	 * ArchiveRequestManager#increaseDeleteTime} festgelegten Wert zu lschen. Das Lschen der Zeitbereiche wird dabei in
	 * zwei Varianten unterteilt. In der ersten Variante werden die Zeitbereiche umgehend aus dem direkten Zugriff des
	 * Archivsystems entfernt. In der zweiten Variante werden die Zeitbereiche nur als "zu lschend" markiert, sobald das
	 * "automatische Lschen" des Archivsystems angstoen wird (dies geschieht zyklisch), werden die Zeitbereiche
	 * entfernt.
	 * <p/>
	 * Bei allen Lschoperationen, die durch diesen Methodenaufruf ausgelst werden, muss darauf geachtete werden, dass der
	 * Vorhaltezeitraum der Zeitbereiche abgelaufen sein muss und das die Zeitbereiche gesichert wurden, falls diese
	 * gesichert werden sollten. Wird gegen eine diese Forderungen verstossen, wird der angegebene Zeitbereich nicht
	 * gelscht.
	 *
	 * @param dataDisposedToDelete Zeitbereich(e), die gelscht werden sollen
	 * @param deleteImmediately    true = Variante 1, die Zeitbereiche werden umgehend aus dem direkten Zugriff des
	 *                             Archivsystems entfernt; false = Variante 2, die Zeitbereiche werden nur als "zu
	 *                             lschend" markiert und spter aus dem direkten Zugriff des Archivsystems entfernt
	 * @return Ergebnisobjekt, ber das Informationen ber den Zustand des Auftrags abgefragt werden knnen
	 */
	ArchiveQueryResult deleteData(List<ArchiveInformationResult> dataDisposedToDelete, boolean deleteImmediately);

	/**
	 * Beauftragt das Archivsystem seine Informationen zu gespeicherten Daten mit einem Speichermedium Typ B {@link
	 * de.bsvrz.dav.daf.main.impl.archive.filesaver.ArchiveFileSaver} abzugleichen. Dies kann ntig werden, wenn die eindeutigen Identifizierungen
	 * der Speichermedien von Typ B durch die Sicherung gendert wurden (Beispiel: Die Daten wurden vorher auf CD
	 * gespeichert, nun werden die Daten auf DVD gespeichert und die alten Datenbestnde auf DVD umkopiert, somit fallen
	 * mehrere CDs auf eine DVD und die eindeutigen Identifizierungen der CDs sind nutzlos. Die eindeutigen
	 * Identifizierungen der CDs wurde aber vom Archivsystem gespeichert und mssen folglich abgeglichen werden).
	 *
	 * @param volumeIdTypB Eindeutige Identifikation eines Speichermediums Typ B
	 * @return Ergebnisobjekt, ber das Informationen ber den Zustand des Auftrags abgefragt werden knnen
	 */
	ArchiveQueryResult archiveFileSaverAlignment(int volumeIdTypB);

	/**
	 * Stellt fest ob das Archivsystem ber den Datenverteiler derzeit erreichbar ist. Eine positive/negative Antwort ist
	 * keine Garantie, dass dieser Zustand auch in Zukunft gltig ist/bleibt.
	 *
	 * @return true = Das Archivsystem ist derzeit erreichbar, es knnen alle im Interface spezifizierten Methoden benutzt
	 *         werden; false = Das Archivsystem ist derzeit nicht erreichbar
	 */
	boolean isArchiveAvailable();

	/**
	 * Das bergebene Objekt <code>listener</code> wird benachrichtigt sobald sich die Erreichbarkeit des Archivsystems
	 * ber den Datenverteiler ndert.
	 *
	 * @param listener Objekt, das benutzt wird um nderungen der Erreichbarkeit des Archivsystems ber den Datenverteiler
	 *                 anzuzeigen
	 */
	void addArchiveAvailabilityListener(ArchiveAvailabilityListener listener);

	/**
	 * Das Objekt, das mit {@link #addArchiveAvailabilityListener} bergeben wurde, wird nicht mehr benachrichtigt sobald
	 * sich die Erreichbarkeit des Archivsystems ber den Datenverteiler ndert.
	 *
	 * @param listener Objekt, das nicht mehr benachrichtigt werden soll, wenn sich die Erreichbarkeit des Archivsystems
	 *                 ber den Datenverteiler ndert
	 */
	void removeArchiveAvailabilityListener(ArchiveAvailabilityListener listener);

	/**
	 * Beauftragt das Archivsystem fehlende Daten von anderen Archivsystemen anzufordern und diese dann als "nachgefordert"
	 * zu speichern und bei Archivanfragen zur Verfgung zu stellen. Sind die bergebenen Listen
	 *
	 * @param requiredData      Datenidentifikation(en), die nachgefordert werden sollen. Jede Datenidentifikation
	 *                          speichert zustlich den Zeitbereich, in dem Daten zu dieser Datenidentifikation angefordert
	 *                          werden soll. Es muss mindestens eine Datenidentifikation vorhanden sein.
	 * @param requestedArchives Archivsystem(e), bei denen Daten angefordert werden. Ist diese Liste leer werden alle
	 *                          Archivsystem angefragt, die beim automatischen Nachfordern angefragt werden.
	 * @return Ergebnisobjekt, ber das Informationen ber den Zustand des Auftrags abgefragt werden knnen
	 */
	ArchiveQueryResult requestData(Collection<ArchiveInformationResult> requiredData, Collection<SystemObject> requestedArchives);

	/**
	 * Beauftragt das Archivsystem fehlende Daten von anderen Archivsystemen anzufordern und diese dann als "nachgefordert"
	 * zu speichern und bei Archivanfragen zur Verfgung zu stellen. Es werden alle Datenidentifikationen angefragt, die
	 * auch beim automatischen Nachfordern angefragt werden.
	 *
	 * @param startTime         Startzeitpunkt, ab dem Daten nachgefordert werden
	 * @param endTime           Endzeitpunkt, bis zu dem Daten nachgefordert werden
	 * @param requestedArchives Archivsysteme die angefragt werden. Ist die Liste leer werden alle Archivsysteme angefragt,
	 *                          die beim automatischen Nachfordern angefragt werden
	 * @return Ergebnisobjekt, ber das Informationen ber den Zustand des Auftrags abgefragt werden knnen
	 */
	ArchiveQueryResult requestData(long startTime, long endTime, Collection<SystemObject> requestedArchives);
}
