/*
 * Copyright 2008 by Kappich Systemberatung, Aachen
 * Copyright 2005 by Kappich+Kni Systemberatung Aachen (K2S)
 * 
 * This file is part of de.bsvrz.dav.daf.
 * 
 * de.bsvrz.dav.daf is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.dav.daf is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with de.bsvrz.dav.daf; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.dav.daf.main.config;

/**
 * Definiert Gemeinsamkeiten aller dynamischen Objekt-Typen. Dynamische Typen haben die Eigenschaft, da beim Erzeugen bzw. Lschen von Objekten dieses Typs
 * diese nderungen sofort gltig werden.
 *
 * @author Kappich Systemberatung
 * @version $Revision: 5982 $
 */
public interface DynamicObjectType extends SystemObjectType, MutableCollection {

	/**
	 * Bestimmt den Persistenzmodus von dynamischen Objekten dieses Typs.
	 *
	 * @return Persistenzmodus von dynamischen Objekten dieses Typs
	 */
	public PersistenceMode getPersistenceMode();

	/**
	 * Fgt einen Listener zu einer Datenstruktur hinzu. Sobald ein neues dynamisches Objekt mit diesem Typ in der Konfiguration erzeugt wird, wird der Listener
	 * informiert.
	 * <p/>
	 * Der mehrfache Aufruf der Methode mit dem identischen Objekt(==) <code>objectCreatedListener</code> fgt das Objekt nur einmal der Datenstruktur hinzu.
	 *
	 * @param objectCreatedListener Objekt, das informiert wird sobald ein dynamisches Objekt mit diesem Typ erzeugt wird.
	 */
	public void addObjectCreationListener(DynamicObjectCreatedListener objectCreatedListener);

	/**
	 * Entfernt den Listener aus der Datenstruktur. Nach Aufruf dieser Methode findet keine Benachrichtigung ber neue Objekte mehr statt.
	 *
	 * @param objectCreatedListener Objekt, das aus der Datenstruktur entfernt werden soll.
	 *
	 * @see #addObjectCreationListener
	 */
	public void removeObjectCreationListener(DynamicObjectCreatedListener objectCreatedListener);

	/**
	 * Fgt einen Listener zu einer Datenstruktur hinzu. Sobald der Name eines dynamischen Objekts dieses Typs gendert wird, wird der Listener informiert.
	 * <p/>
	 * Ein mehrfacher Aufruf dieser Methode mit dem selben Objekten fgt das bergene Objekt nur einmal der Datenstruktur hinzu.
	 *
	 * @param nameChangeListener Objekt, das informiert wird, sobald der Name eines Objekts von diesem Typ gender wird.
	 */
	void addNameChangeListener(NameChangeListener nameChangeListener);

	/**
	 * Entfernt den Listener aus der Datenstruktur. Das Objekt wird nicht mehr informiert wenn sich der Name eines Objekts von diesem Typ ndert.
	 *
	 * @param nameChangeListener Listener, der entfernt wird.
	 *
	 * @see #addNameChangeListener
	 */
	void removeNameChangeListener(NameChangeListener nameChangeListener);

	/**
	 * Fgt einen Listener zu einer Datenstruktur hinzu. Sobald ein dynamisches Objekt dieses Typs ungltig wird, wird der Listener informiert.
	 * <p/>
	 * Ein mehrfacher Aufruf dieser Methode mit dem selben Objekten fgt das bergene Objekt nur einmal der Datenstruktur hinzu.
	 *
	 * @param invalidationListener Objekt, das informiert wird, sobald ein Objekt des Typs ungltig wird.
	 */
	void addInvalidationListener(InvalidationListener invalidationListener);

	/**
	 * Entfernt den Listener aus der Datenstruktur. Das Objekt wird nicht mehr informiert wenn ein Objekt dieses Typs ungltig wird.
	 *
	 * @param invalidationListener Listener, der entfernt werden soll.
	 */
	void removeInvalidationListener(InvalidationListener invalidationListener);


	/** Aufzhlung fr die verschiedenen Persistenzmodi der dynamischen Objekte eines Typs */
	enum PersistenceMode {

		/** Dynamische Objekte werden nicht persistent gespeichert. */
		TRANSIENT_OBJECTS,

		/** Dynamische Objekte werden persistent gespeichert. */
		PERSISTENT_OBJECTS,

		/** Dynamische Objekte werden persistent gespeichert und beim Neustart auf ungltig gesetzt. */
		PERSISTENT_AND_INVALID_ON_RESTART, }

	/** Listener, der eine Methode zur Verfgung stellt, die aufgerufen wird sobald sich der Name eines Objekts ndert. */
	public static interface NameChangeListener {

		/**
		 * Wird aufgerufen, wenn sich der Name eines dynamischen Objekts ndert.
		 *
		 * @param newObject dynamisches Objekte, dessen Name sich gendert hat.
		 */
		void nameChanged(DynamicObject newObject);
	}

	/** Listener, dessen Methode benutzt wird wenn ein neues dynamisches Objekt angelegt wurde. */
	public static interface DynamicObjectCreatedListener {

		/**
		 * Diese Methode wird aufgerufen, wenn ein neues Objekt erstellt wurde.
		 *
		 * @param createdObject Objekt, das neu erstellt wurde.
		 */
		public void objectCreated(DynamicObject createdObject);
	}
}

