/*
 * Copyright 2008 by Kappich Systemberatung, Aachen
 * Copyright 2006 by Kappich Systemberatung Aachen
 * 
 * This file is part of de.bsvrz.dav.daf.
 * 
 * de.bsvrz.dav.daf is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.dav.daf is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with de.bsvrz.dav.daf; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.dav.daf.main.impl.config.request.telegramManager;

import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.config.MutableCollectionChangeListener;
import de.bsvrz.dav.daf.main.impl.config.request.RequestException;

/**
 * Dieses Interface emglicht eine Kommunikation mit einem Sender und einem Empfnger. Der Sender verschickt Auftrge und empfngt dann die Antworten auf diese
 * Auftrge.
 *
 * @author Kappich Systemberatung
 * @version $Revision: 5982 $
 */
public interface SenderReceiverCommunication {

	/**
	 * Verschickt eine Anfrage vom angegeben Typ und mit dem in einem Byte Array angegebenen serialisierten Inhalt, die Methode gibt ohne zu blockieren eine
	 * neue Anfragenummer zurck. Die Antwort kann mit {@link #waitForReply} abgerufen werden.
	 *
	 * @param messageType Anfragetyp
	 * @param data Daten Serialiserte Anfragedaten.
	 *
	 * @return Index, der bentigt wird um die Antwort auf eine Anfrage zu abzufragen.
	 *
	 * @throws SendSubscriptionNotConfirmed Wenn noch keine positive Sendesteuerung vom Datenverteiler fr die zu versendenden Daten vorliegt
	 * @throws IllegalStateException        Die Daten knnen versendet werden, aber die Antwort kann nicht empfangen werden, da fr den empfang der Daten nicht die
	 *                                      bentigten Rechte vorhanden sind. Aus diesem Grund werden die Daten nicht verschickt.
	 */
	public int sendData(String messageType, byte[] data) throws SendSubscriptionNotConfirmed, IllegalStateException;

	/**
	 * Verschickt eine Antwort vom angegeben Typ zu einer Anfrage.
	 *
	 * @param messageType Antworttyp
	 * @param data Daten Serialiserte Anfragedaten.
	 * @param queryIndex Index der zugehrigen Anfrage
	 *
	 * @throws SendSubscriptionNotConfirmed Wenn noch keine positive Sendesteuerung vom Datenverteiler fr die zu versendenden Daten vorliegt
	 * @throws IllegalStateException        Die Daten knnen versendet werden, aber die Antwort kann nicht empfangen werden, da fr den empfang der Daten nicht die
	 *                                      bentigten Rechte vorhanden sind. Aus diesem Grund werden die Daten nicht verschickt.
	 */
	public void sendData(String messageType, byte[] data, int queryIndex) throws SendSubscriptionNotConfirmed, IllegalStateException;

	/**
	 * Stellt die Antwort auf eine Anfrage zur Verfgung
	 *
	 * @param reqestIndex Index, der bei der Methode {@link #sendData} als Rckgabeparameter zurckgegeben wurde
	 *
	 * @return Antwort auf eine Anfrage
	 * @throws de.bsvrz.dav.daf.main.impl.config.request.RequestException Wenn die Kommunikation zum Datenverteiler unterbrochen wurde.
	 */
	public Data waitForReply(int reqestIndex) throws RequestException;

	/** Schliet alle geffneten Verbindungen und beendet mgliche Threads */
	public void close();

	/**
	 * Diese Methode gibt den derzeitigen Zustand einer Anmeldung als Sender/Quelle wieder.
	 *
	 * @return derzeitiger Zustand der Verbindung
	 */
	public ConnectionState getConnectionState();

	/**
	 * Setzt den Listener zur Verarbeitung und Verteilung von Aktualisierungsnachrichten bzgl. nderungen der Elemente von dynamischen Mengen bzw. dynamischen
	 * Typen.
	 * @param notifyingMutableCollectionChangeListener Listener zur Verarbeitung und Verteilung von Aktualisierungsnachrichten.
	 */
	void setMutableCollectionChangeListener(final MutableCollectionChangeListener notifyingMutableCollectionChangeListener);

	/** Diese Klasse spiegelt die Verbindung einer Sende/Quelle-Anmeldung wieder. */
	public final static class ConnectionState {

		public final static ConnectionState NotYetConnected = new ConnectionState("Noch nicht verbunden");

		public final static ConnectionState Connected = new ConnectionState("Verbunden");

		public final static ConnectionState Disconnected = new ConnectionState("Nicht mehr verbunden");

		public final static ConnectionState Error = new ConnectionState("Fehler");

		public String toString() {
			return _name;
		}

		private String _name;

		private ConnectionState(String name) {
			_name = name;
		}
	}
}
