/*
 * Copyright 2007 by Kappich Systemberatung, Aachen
 * Copyright 2004 by Kappich+Kni Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.dav.dav.
 * 
 * de.bsvrz.dav.dav is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.dav.dav is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.dav.dav; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.dav.dav.main;

import de.bsvrz.dav.daf.communication.dataRepresentation.AttributeBaseValue;
import de.bsvrz.dav.daf.communication.dataRepresentation.AttributeListValue;
import de.bsvrz.dav.daf.communication.dataRepresentation.AttributeValue;
import de.bsvrz.dav.daf.communication.dataRepresentation.datavalue.ByteAttribute;
import de.bsvrz.dav.daf.communication.dataRepresentation.datavalue.LongAttribute;
import de.bsvrz.dav.daf.communication.lowLevel.telegrams.BaseSubscriptionInfo;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.AttributeGroupUsage;
import de.bsvrz.dav.daf.main.config.ConfigurationException;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.ObjectSet;
import de.bsvrz.dav.daf.main.config.ObjectSetType;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.SystemObjectType;
import de.bsvrz.dav.daf.main.config.ConfigurationArea;
import de.bsvrz.dav.daf.main.config.ObjectTimeSpecification;
import de.bsvrz.dav.daf.main.impl.config.DafDataModel;
import de.bsvrz.sys.funclib.debug.Debug;

import java.util.*;

/**
 * Verwaltet die Rechte eines Benutzers.
 *
 * @author Kappich Systemberatung
 * @version $Revision: 6236 $
 */
public class UserInfo implements ClientReceiverInterface {

	private static final Debug _debug = Debug.getLogger();

	private static final String USER_ATTRIBUTE_GROUP_PID = "atg.benutzerParameter";

	private static final String USER_ASPECT_PID = "asp.parameterSoll";

	private static final String AUTHENTIFICATION_CLASS_ATTRIBUTE_GROUP_PID = "atg.rollenRegionenPaareParameter";

	private static final String AUTHENTIFICATION_CLASS_ASPECT_PID = "asp.parameterSoll";

	private static final String ROLE_ACTIVITIES_SET_NAME = "Aktivitten";

	private static final String ACTIVITY_ATTRIBUTE_GROUP_SET_NAME = "Attributgruppen";

	private static final String ACTIVITY_ASPECT_SET_NAME = "Aspekte";

	private static final String ACTIVITY_RIGHTS_ATTRIBUTE_GROUP_PID = "atg.zugriffsRechte";

	private static final String REGION_OBJECT_SET_NAME = "Objekte";

	private static final String REGION_CONTAINER_SET_NAME = "Zusammenstellungen";

	/**
	 * Damit die Software auch mit lteren Versionen des Datenmodells arbeitet, wird hier neben dem
	 * korrekten Namen der Menge auch der ursprngliche, falsch geschriebene Name der Menge untersttzt.
	 */
	private static final String REGION_CONTAINER_SET_BAD_NAME = "Zusammmenstellungen";

	private static final String REGION_TYPE_PID = "typ.zugriffsRegion";

	private static final String CONFIGURATION_SUBSET_TYPE_PID = "typ.konfigurationsBereich";

	private static final String CONFIGURATION_SUBSET_OBJECTS_SET_NAME = "Objekte";

	/** Id des Benutzers */
	private long _userId;

	/** Der Verbindungsmanager */
	ConnectionsManager connectionsManager;

	/** Stellt alle Rollen/Regionen-Kombinationen dar, die fr diesen Benutzer erlaubt sind. */
	private ArrayList _authentificationUnits;

	/** Datenverteilerobjekt, das den Benutzer darstellt. */
	private SystemObject _user;

	/** Beschreibung der Benutzeranmeldedaten */
	private DataDescription _userDataDescription;

	/** Eine interne Klasse zur Aktuallisierung der Benutzerrechten */
	private AuthentificationClassUpdater _updater;

	/** Flage zur Bereitschaftstatus */
	private boolean _initComplete = false;

	/** Erster Durchlauf */
	private boolean _firstTime = true;

	/** Zhlt wie oft ein UserInfo-Objekt angelegt werden sollte, aber es konnte die Referenz auf das bereits bestehende Objekt benutzt werden. */
	private int _reference = 1;

	/**
	 * @param userId             Id des Benutzers. ber die Id wird das Objekt des Benutzers vom Datenverteiler angefordert.
	 * @param _connectionsManager Wird zum an/abmelden von Daten benutzt.
	 */
	public UserInfo(long userId, ConnectionsManager _connectionsManager) {
		if((userId == -1) || (_connectionsManager == null)) {
			throw new IllegalArgumentException("Argument ist null");
		}
		_initComplete = false;
		_userId = userId;
		connectionsManager = _connectionsManager;
		_authentificationUnits = new ArrayList();
		try {
			_user = (SystemObject)connectionsManager.dataModel.getObject(_userId);
			if(_user == null) {
				return;
			}
			AttributeGroup attributeGroup = (AttributeGroup)connectionsManager.dataModel.getObject(USER_ATTRIBUTE_GROUP_PID);
			if(attributeGroup == null) {
				return;
			}
			Aspect aspect = (Aspect)connectionsManager.dataModel.getObject(USER_ASPECT_PID);
			if(aspect == null) {
				return;
			}
			_userDataDescription = new DataDescription(attributeGroup, aspect);
			connectionsManager.selfClientDavConnection.subscribeReceiver(
					this, _user, _userDataDescription, ReceiveOptions.normal(), ReceiverRole.receiver()
			);
		}
		catch(ConfigurationException ex) {
			ex.printStackTrace();
			return;
		}
	}

	/**
	 * Id, die das Systemobjekt des Benutzers darstellt.
	 * @return Id, mit der das Systemobjekt des Benutzers angefordert werden kann.
	 */
	public final long getUserId() {
		return _userId;
	}

	/**
	 * Der Benutzer sollte erneut angemeldet werden.
	 */
	public final void incrementReference() {
		++_reference;
	}

	/**
	 * Der Benutzer sollte einmal abgemeldet werden
	 */
	public final void decrementReference() {
		--_reference;
	}

	/**
	 * Der Benutzer ist nicht mehr angemeldet.
	 * @return <code>true</code>, wenn der Benutzer vom System abgemeldet wurde.
	 */
	public final boolean canBeSafelyDeleted() {
		return _reference == 0;
	}

	/** Wurde fr automatisierte Tests hinzugefgt. */
	int getNumberOfReferences() {
		return _reference;
	}

	/**
	 * Wird fr automatisierte Tests bentigt.
	 * @return Objekt, mit dem Daten fr die die ATG atg.rollenRegionenPaareParameter bergeben werden kann. 
	 */
	ClientReceiverInterface getClassUpdater(){
		return _updater;
	}

	/**
	 * Aktualisierungsmethode, die nach Empfang eines angemeldeten Datensatzes von den Datenverteiler-Applikationsfunktionen aufgerufen wird. Diese Methode muss
	 * von der Applikation zur Verarbeitung der empfangenen Datenstze implementiert werden.
	 *
	 * @param results Feld mit den empfangenen Ergebnisdatenstzen.
	 */
	public final void update(ResultData results[]) {
		if(results == null) {
			return;
		}
		ResultData result = null;
		long dataIndex = -1;
		for(int i = 0; i < results.length; ++i) {

			ResultData resultData = results[i];

			if(resultData == null) {
				continue;
			}
			if(!resultData.getObject().equals(_user)) {
				continue;
			}
			if(!resultData.getDataDescription().getAttributeGroup().equals(_userDataDescription.getAttributeGroup())) {
				continue;
			}
			if(!resultData.getDataDescription().getAspect().equals(_userDataDescription.getAspect())) {
				continue;
			}
			if(!resultData.hasData()) {
				continue;
			}
			if(resultData.getDataIndex() > dataIndex) {
				result = resultData;
			}
		}
		if(result != null) {
			final Data data = result.getData();
			final ConfigurationObject authentificationClass = (ConfigurationObject)data.getReferenceValue("berechtigungsklasse").getSystemObject();
			if(authentificationClass != null) {

				_initComplete = false;
				// Dieser Thread meldet nur einen Empfnger an und wird dann beendet
				Runnable runnable = new Runnable() {
					public final void run() {
						if(_updater == null) {
							_initComplete = false;
							_updater = new AuthentificationClassUpdater(authentificationClass);
						}
						else {
							if(!authentificationClass.equals(_updater.getAuthentificationClass())) {
								_initComplete = false;
								_updater.disable();
								_updater = new AuthentificationClassUpdater(authentificationClass);
							}
							else {
								_initComplete = true;
							}
						}
					}
				};
				(new Thread(runnable)).start();
			}
		}
	}

	/**
	 * Prft, ob die angegebenen Parameter in einer Berechtigungsklasse festgelegt sind. Sind die Parameter in einer Berechtigungsklasse zu finden, so drfen die
	 * Daten beim DaV angemeldet werden.
	 *
	 * @param info  Daten, die angemeldet werden sollen.
	 * @param state <ul><li>state 0: Als Sender</li> <li>1: Als Empfnger</li> <li>2: Als Quelle</li> <li>3: Als Senke</li></ul>
	 *
	 * @return <code>true</code>, wenn die Daten angemeldet werden drfen, sonst <code>false</code>.
	 */
	public final boolean authentify(BaseSubscriptionInfo info, byte state) {
		synchronized(connectionsManager) {
			while(!_initComplete) {
				try {
					connectionsManager.wait();
				}
				catch(InterruptedException ex) {
					break;
				}
			}
		}
		Long id = new Long(info.getObjectID());
		AttributeGroupUsage atgUsage = connectionsManager.dataModel.getAttributeGroupUsage(info.getUsageIdentification());
		if(atgUsage == null) return false;
		Long attributeGroupId = new Long(atgUsage.getAttributeGroup().getId());
		Long aspectId = new Long(atgUsage.getAspect().getId());
		for(int i = _authentificationUnits.size() - 1; i > -1; --i) {
			AuthentificationUnit unit = (AuthentificationUnit)_authentificationUnits.get(i);
			if(unit != null) {
				if(unit.isAllowed(id, attributeGroupId, aspectId, state)) {
					return true;
				}
			}
		}
		return false;
	}

	/** Meldet die Empfnger fr die Daten atg.benutzerParameter und atg.rollenRegionenPaareParameter ab. */
	public final void disable() {
		try {
			connectionsManager.selfClientDavConnection.unsubscribeReceiver(
					this, _user, _userDataDescription
			);
			if(_updater != null) {
				_updater.disable();
			}
		}
		catch(ConfigurationException ex) {
			ex.printStackTrace();
			return;
		}
	}

	public final int hashCode() {
		int result = 19;
		long id = _user.getId();
		result = (11 * result) + (int)(id ^ (id >>> 32));
		return result;
	}

	public final boolean equals(Object obj) {
		if(obj == null) {
			return false;
		}
		if(obj instanceof UserInfo) {
			UserInfo ui = (UserInfo)obj;
			if(_user.getId() == ui._user.getId()) {
				return true;
			}
		}
		return false;
	}

	private final Activity[] getRoleActivities(ConfigurationObject role) {
		Activity activities[] = null;
		if(role != null) {
			ObjectSet activitiesSet = role.getObjectSet(ROLE_ACTIVITIES_SET_NAME);
			if(activitiesSet != null) {
				List activitiesList = activitiesSet.getElements();
				if(activitiesList != null) {
					int size = activitiesList.size();
					if(size > 0) {
						activities = new Activity[size];
						for(int i = 0; i < size; ++i) {
							ConfigurationObject activity = (ConfigurationObject)activitiesList.get(i);
							if(activity != null) {
								ObjectSet attributeGroupSet = activity.getObjectSet(ACTIVITY_ATTRIBUTE_GROUP_SET_NAME);
								ArrayList attributegroups = new ArrayList();
								if(attributeGroupSet != null) {
									List atgs = attributeGroupSet.getElements();
									if(atgs != null) {
										for(int j = atgs.size() - 1; j > -1; --j) {
											AttributeGroup atg = (AttributeGroup)atgs.get(j);
											if(atg != null) {
												attributegroups.add(new Long(atg.getId()));
											}
										}
									}
								}
								ObjectSet aspectSet = activity.getObjectSet(ACTIVITY_ASPECT_SET_NAME);
								ArrayList aspects = new ArrayList();
								if(aspectSet != null) {
									List asps = aspectSet.getElements();
									if(asps != null) {
										for(int j = asps.size() - 1; j > -1; --j) {
											Aspect asp = (Aspect)asps.get(j);
											if(asp != null) {
												aspects.add(new Long(asp.getId()));
											}
										}
									}
								}
								byte mode = 0;
								AttributeGroup rightsAttributeGroup = (AttributeGroup)connectionsManager.dataModel.getObject(ACTIVITY_RIGHTS_ATTRIBUTE_GROUP_PID);
								List rightsValues = ((DafDataModel)connectionsManager.dataModel).getObjectDataValues(
										activity, rightsAttributeGroup
								);
								if(rightsValues != null) {
									byte read = ((Byte)((ByteAttribute)((AttributeValue)rightsValues.get(0)).getValue()).getValue()).byteValue();
									byte write = ((Byte)((ByteAttribute)((AttributeValue)rightsValues.get(1)).getValue()).getValue()).byteValue();
									byte main = ((Byte)((ByteAttribute)((AttributeValue)rightsValues.get(2)).getValue()).getValue()).byteValue();
									if(read == 1) {
										mode |= 0x01;
									}
									if(write == 1) {
										mode |= 0x02;
									}
									if(main == 1) {
										mode |= 0x04;
									}
									activities[i] = new Activity(attributegroups, aspects, mode);
								}
							}
						}
					}
				}
			}
		}
		return activities;
	}

	private final void collectIds(List ids, List objects) {
		if(objects == null) {
			return;
		}
		int size = objects.size();
		for(int i = 0; i < size; ++i) {
			SystemObject obj = (SystemObject)objects.get(i);
			if(obj != null) {
				ids.add(new Long(obj.getId()));
			}
		}
	}

	private final ArrayList getRegionObjects(ConfigurationObject region) throws ConfigurationException {
		ArrayList objectList = new ArrayList();
		if(region != null) {
			ObjectSet objectsSet = region.getObjectSet(REGION_OBJECT_SET_NAME);
			if(objectsSet != null) {
				collectIds(objectList, objectsSet.getElements());
			}
			ObjectSet setsSet = region.getObjectSet(REGION_CONTAINER_SET_NAME);
			if(setsSet == null) {
				// Damit die Software auch mit lteren Versionen des Datenmodells arbeitet, wird hier neben dem
				// korrekten Namen der Menge auch der ursprngliche, falsch geschriebene Name der Menge untersttzt.
				setsSet = region.getObjectSet(REGION_CONTAINER_SET_BAD_NAME);
				if(setsSet != null) {
					_debug.warning("In der Region " + region + " wird die Menge 'Zusammmenstellungen' mit 3 'm' verwendet. Bitte kb.systemModellGlobal "
					               + "aktualisieren und Mengenname ndern! ");
				}
			}
			if(setsSet != null) {
				List tmp = setsSet.getElements();
				if((tmp != null) && (tmp.size() > 0)) {
					for(int i = tmp.size() - 1; i > -1; --i) {
						SystemObject o = (SystemObject)tmp.get(i);
						if(o != null) {
							collectIds(objectList, getContainerObjects(o));
						}
					}
				}
			}
		}
		return objectList;
	}

	private final List getContainerObjects(SystemObject container) throws ConfigurationException {
		if(container != null) {
			SystemObjectType type = container.getType();
			if(type instanceof ObjectSetType) {
				return ((ObjectSet)container).getElements();
			}
			if(REGION_TYPE_PID.equals(type.getPid())) {
				return getRegionObjects((ConfigurationObject)container);
			}
			if(CONFIGURATION_SUBSET_TYPE_PID.equals(type.getPid())) {
				return getConfigurationSubsetObjects((ConfigurationObject)container);
			}
		}
		return new ArrayList();
	}

	private final List getConfigurationSubsetObjects(SystemObject container) throws ConfigurationException {
		List objectList = new ArrayList();
    	if(container instanceof ConfigurationArea) {
		    ConfigurationArea area = (ConfigurationArea)container;
		    final ArrayList<ConfigurationArea> areas = new ArrayList<ConfigurationArea>();
		    areas.add(area);
		    objectList.addAll(container.getDataModel().getObjects(areas, null, ObjectTimeSpecification.valid()));
		}
		return objectList;
	}

	private class AuthentificationClassUpdater implements ClientReceiverInterface {

		/** Die Authentificationsklasse */
		private SystemObject authentificationClass;

		/** Beschreibung der Authentifikationsanmeldedaten */
		private DataDescription authentificationClassDataDescription;

		/**
		 * Meldet sich als Empfnger auf das Objekt an, das die Berechtigungsklasse fr den Benutzer darstellt. Sobald es nderungen gibt (Regionen und/oder Aktionen
		 * ndern sich), werden diese nderungen am Java-Objekt, das die Berechtigungsklasse darstellt, bernommen.
		 *
		 * @param _authentificationClass Datenverteilerobjekt, das eine Berechtigungsklasse darstellt.
		 */
		public AuthentificationClassUpdater(SystemObject _authentificationClass) {
			try {
				authentificationClass = _authentificationClass;
				AttributeGroup attributeGroup = (AttributeGroup)connectionsManager.dataModel.getObject(AUTHENTIFICATION_CLASS_ATTRIBUTE_GROUP_PID);
				if(attributeGroup == null) {
					return;
				}
				Aspect aspect = (Aspect)connectionsManager.dataModel.getObject(AUTHENTIFICATION_CLASS_ASPECT_PID);
				if(aspect == null) {
					return;
				}
				authentificationClassDataDescription = new DataDescription(attributeGroup, aspect);
				connectionsManager.selfClientDavConnection.subscribeReceiver(
						this, authentificationClass, authentificationClassDataDescription, ReceiveOptions.normal(), ReceiverRole.receiver()
				);
			}
			catch(ConfigurationException ex) {
				ex.printStackTrace();
				return;
			}
		}

		public final SystemObject getAuthentificationClass() {
			return authentificationClass;
		}

		/** Meldet sich als Empfnger von nderungen auf Regionen/Aktionen ab. */
		public final void disable() {
			try {
				connectionsManager.selfClientDavConnection.unsubscribeReceiver(
						this, authentificationClass, authentificationClassDataDescription
				);
			}
			catch(ConfigurationException ex) {
				ex.printStackTrace();
				return;
			}
		}

		/**
		 * Aktualisierungsmethode, die nach Empfang eines angemeldeten Datensatzes von den Datenverteiler-Applikationsfunktionen aufgerufen wird. Diese Methode muss
		 * von der Applikation zur Verarbeitung der empfangenen Datenstze implementiert werden.
		 *
		 * @param results Feld mit den empfangenen Ergebnisdatenstzen.
		 */
		public void update(ResultData[] results) {
			if(results == null) {
				return;
			}
			try {
				ResultData result = null;
				long dataIndex = -1;
				for(int i = 0; i < results.length; ++i) {
					ResultData resultData = results[i];
					if(resultData == null) {
						continue;
					}
					if(!resultData.getObject().equals(authentificationClass)) {
						continue;
					}
					if(!resultData.getDataDescription().getAttributeGroup().equals(authentificationClassDataDescription.getAttributeGroup())) {
						continue;
					}
					if(!resultData.getDataDescription().getAspect().equals(authentificationClassDataDescription.getAspect())) {
						continue;
					}
					if(!resultData.hasData()) {
						continue;
					}
					if(resultData.getDataIndex() > dataIndex) {
						result = resultData;
					}
				}
				if(result != null) {
					_authentificationUnits.clear();
					final Data data = result.getData();
					_debug.fine("data = " + data);
					final Data.Array roleRegionPairs = data.getArray("rollenRegionenPaare");
					final int numberOfPairs = roleRegionPairs.getLength();
					for(int i = 0; i< numberOfPairs; ++i) {
						final Data roleRegionPair = roleRegionPairs.getItem(i);
						ConfigurationObject role = (ConfigurationObject)roleRegionPair.getReferenceValue("rolle").getSystemObject();
						ConfigurationObject region = (ConfigurationObject)roleRegionPair.getReferenceValue("region").getSystemObject();
						if((role != null) && (region != null)) {
							Region regionWrapper = new Region(getRegionObjects(region));
							Activity activities[] = getRoleActivities(role);
							AuthentificationUnit unit = new AuthentificationUnit(regionWrapper, activities);
							_authentificationUnits.add(unit);
						}
					}
				}
			}
			catch(ConfigurationException ex) {
				ex.printStackTrace();
			}
			finally {
				synchronized(connectionsManager) {
					_initComplete = true;
					connectionsManager.notifyAll();
					if(_firstTime) {
						_firstTime = false;
					}
					else {
						try {
							connectionsManager.handleUserRightsChanged(_userId);
						}
						catch(Exception e) {
							_debug.warning("Fehler beim ndern der Zugriffsrechte des Benutzers " + _user, e);

						}
						finally {
							_debug.info("Zugriffsrechte des Benutzers " + _user + " haben sich gendert.");

						}
					}
				}
			}
		}

		public final int hashCode() {
			int result = 19;
			long id = _user.getId();
			result = (41 * result) + (int)(id ^ (id >>> 32));
			id = authentificationClass.getId();
			result = (41 * result) + (int)(id ^ (id >>> 32));
			return result;
		}

		public final boolean equals(Object obj) {
			if(obj == null) {
				return false;
			}
			if(obj instanceof AuthentificationClassUpdater) {
				AuthentificationClassUpdater au = (AuthentificationClassUpdater)obj;
				if((getAssociatedUserId() == au.getAssociatedUserId()) && (authentificationClass.getId() == au.authentificationClass.getId())) {
					return true;
				}
			}
			return false;
		}

		public final long getAssociatedUserId() {
			return _user.getId();
		}
	}

	/**
	 * Speichert unterschiedliche Objekte (ATG, Aspekte, Objekte) und stellt eine Methode zur Verfgung, mit der geprft werden kann, ob ein bestimmtes Objekt
	 * vorhanden ist.
	 */
	private class InfoHolder {

		/** true, wenn die im Konstruktor bergebene Liste null oder leer ist; false, sonst. */
		private boolean all;

		// Diese Variable sollte durch ein HashSet ersetzt werden, da nur ein "Contains" bentigt wird.
		/** Speichert die Objekte der im Konstruktor bergebenen Liste. Als Schlssel dient das Objekt, als Value wird ebenfalls das Objekt eingetragen. */
		private Hashtable table;

		public InfoHolder(ArrayList list) {
			if((list == null) || (list.size() == 0)) {
				all = true;
			}
			else {
				all = false;
				table = new Hashtable();
				for(int i = list.size() - 1; i > -1; --i) {
					Object o = list.get(i);
					if(o != null) {
						table.put(o, o);
					}
				}
			}
		}

		/**
		 * @param o Objekt, das vorhanden sein soll. Wurde im Konstruktor eine leere Liste bzw. <code>null</code> bergeben.
		 *
		 * @return true, wenn a) Das bergebene Objekt in der Liste vorhanden ist, die im Konstruktor bergeben wurde. b) Im Konstruktor eine leere Liste oder
		 *         <code>null</code> bergeben wurde. In allen anderen Fllen wird <code>false</code> zurckgegeben.
		 */
		public final boolean isAllowed(Object o) {
			if(o == null) {
				return false;
			}
			if(all) {
				return true;
			}
			return (table.get(o) != null);
		}
	}

	/**
	 * Diese Klasse stellt eine Rolle/Aktivitt dar. Es wird eine Methode zur Verfgung gestellt, mit der geprft werden kann, ob eine bergebene Rolle/Aktivitt
	 * erlaubt ist.
	 */
	private class Activity {

		/** read  ==> bit index 0 is set (mode & 0x01) write ==> bit index 1 is set (mode & 0x02) main  ==> bit index 2 is set (mode & 0x04) */
		private final byte _mode;

		/** Alle erlaubten ATGs dieser Rolle. */
		private final InfoHolder _attributeGroups;

		/** Alle erlaubten Aspekte dieser Rolle. */
		private final InfoHolder _aspects;

		/**
		 * @param atgs  Alle ATGs, die in dieser Rolle erlaubt sind. Sollen alle ATGs erlaubt sein, so kann <code>null</code> oder eine leere Liste bergeben
		 *              werden.
		 * @param asps  Alle erlaubten Aspekte dieser Rolle. Sollen alle Aspekte erlaubt sein, so kann <code>null</code> oder eine leere Liste bergeben werden.
		 * @param _mode 0: Als Sender 1: Als Empfnger 2: Als Quelle 3: Als Senke
		 */
		public Activity(ArrayList atgs, ArrayList asps, byte _mode) {
			_attributeGroups = new InfoHolder(atgs);
			_aspects = new InfoHolder(asps);
			this._mode = _mode;
		}

		/**
		 * Prft, ob die bergebenen Parameter mit der Rolle/Aktivitt erlaubt sind.
		 *
		 * @param atg   ATG
		 * @param asp   Aspekt
		 * @param mode 0: Als Sender 1: Als Empfnger 2: Als Quelle 3: Als Senke
		 *
		 * @return <code>true</code>, wenn die bergenen Parameter mit der Rolle/Aktivitt erlaubt sind, sonst <code>false</code>.
		 */
		public final boolean isAllowed(Object atg, Object asp, byte mode) {
			switch(mode) {
				case 0: {
					if((_mode & 2) != 2) {
						return false;
					}
					break;
				}
				case 1: {
					if((_mode & 1) != 1) {
						return false;
					}
					break;
				}
				case 2: {
					if((_mode & 6) != 6) {
						return false;
					}
					break;
				}
				case 3: {
					if((_mode & 5) != 5) {
						return false;
					}
					break;
				}
				default: {
					return false;
				}
			}
			return (_attributeGroups.isAllowed(atg) && _aspects.isAllowed(asp));
		}
	}

	/** Diese Klasse stellt eine Region dar und stellt eine Methode zur Verfgung, mit der die Zugehrigkeit anderer Objekte zu der Region geprft werden kann. */
	private class Region {

		private final InfoHolder _objects;

		/**
		 * Objekte, die eine Region definieren.
		 *
		 * @param objs Alle Objekte, die zu einer Region gehren. Eine leere Liste oder <code>null</code> wird als "Wildcard" interpretiert (alles ist erlaubt).
		 */
		public Region(ArrayList objs) {
			_objects = new InfoHolder(objs);
		}

		/**
		 * Diese Methode prft, ob ein Objekt zu einer Region gehrt.
		 *
		 * @param obj Objekt, das Teil einer Region sein soll.
		 *
		 * @return <code>true</code>, wenn das bergebene Objekt in der Liste, die im Konstruktor bergeben wurde, zu finden ist. Wurde im Konstruktor eine leere
		 *         Liste oder <code>null</code> bergeben, wird immer <code>true</code> zurckgegeben. In allen anderen Fllen wird <code>false</code> zurckgegeben.
		 */
		public final boolean isAllowed(Object obj) {
			return _objects.isAllowed(obj);
		}
	}

	/**
	 * Diese Klasse stellt eine Rollen/Regionen-Paar Kombination dar. Sie stellt eine Methode zur Verfgung, mit der geprft werden kann ob ein Objekt, ein
	 * ATG+Aspekt und ein Anmeldung als Senke/Quelle/Empfnger/Sender erlaubt ist.
	 */
	private class AuthentificationUnit {

		/** Region */
		private Region _region;

		/** Rolle/Aktivitt */
		private Activity _activities[];

		/**
		 * @param region     Region
		 * @param activities Alle Aktivitten/Rollen. Wird eine leere Liste bergeben, wird {@link #isAllowed} immer <code>false</code> zurck geben.
		 */
		public AuthentificationUnit(Region region, Activity activities[]) {
			if(region == null) {
				throw new IllegalArgumentException("Region Argument ist null");
			}
			if(activities == null) {
				throw new IllegalArgumentException("Aktivitten Argument ist null");
			}
			_region = region;
			_activities = activities;
		}

		/**
		 * Prft, ob die bergenen Parameter in dieser Berechtigungsklasser erlaubt sind.
		 *
		 * @param obj  Systemobjekt
		 * @param atg  Attributgruppe
		 * @param asp  Aspekt
		 * @param mode 0: Als Sender 1: Als Empfnger 2: Als Quelle 3: Als Senke
		 *
		 * @return <code>true</code>, wenn die bergebenen Parameter mit dieser Berechtigungsklasse erlaubt sind, sonst <code>false</code>.
		 */
		public final boolean isAllowed(Object obj, Object atg, Object asp, byte mode) {
			if(!_region.isAllowed(obj)) {
				return false;
			}
			for(int i = 0; i < _activities.length; ++i) {
				if(_activities[i].isAllowed(atg, asp, mode)) {
					return true;
				}
			}
			return false;
		}
	}
}
