/*
 * Copyright 2008 by Kappich Systemberatung Aachen
 * 
 * This file is part of de.bsvrz.kex.tls.osi2osi3.
 * 
 * de.bsvrz.kex.tls.osi2osi3 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.kex.tls.osi2osi3 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.kex.tls.osi2osi3; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.kex.tls.osi2osi3.redirection;

import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.SystemObjectType;
import de.bsvrz.sys.funclib.debug.Debug;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Klasse TlsModel, die die Informationen zu den Tls-Gerten verwaltet.
 * 
 * @author Kappich Systemberatung
 * @version $Revision: 6171 $
 */
public class TlsModel {
	
	/**
	 * Datenmodell, aus dem die Konfigurationsinformationen bezogen werden
	 */
	private final DataModel _configuration;
	
	/**
	 * Map, in der zu dem Schlssel Systemobjekt die entsprechenden Tls-Knoten vorgehalten werden.
	 */
	private final Map<SystemObject, TlsNode> _tlsDevices;
	
	private static final Debug _debug;
	static {
		_debug = Debug.getLogger();
		//_debug.setLoggerLevel(Debug.FINEST);
	}
	
	/**
	 * Map, die die vorhandenen DE-Typen (PID des DE-Typs) den entsprechenden Funktionsgruppen zuordnet.
	 */
	public Map<String, Integer> _deTyp2FgNr;
	
	/**
	 * Konstruktor. Das bergebene Datenmodell darf nicht null sein. Im Konstruktor werden die Informationen zu den TLS-Gerten aus der Konfiguration gelesen
	 * und in einer eigene Struktur vorgehalten.
	 * 
	 * @param configuration
	 *            Datenmodell
	 */
	public TlsModel(final DataModel configuration) {
		if(configuration == null) throw new IllegalArgumentException("DataModel ist null");
		_configuration = configuration;
		
		_tlsDevices = new HashMap<SystemObject, TlsNode>();
		
		// Aufbau der Map
		_deTyp2FgNr = new HashMap<String, Integer>();
		
		// An dieser Stelle mssen alle der Konfiguration bekannten (Haupt)DE-Typen aufgefhrt sein.
		// Eventuell abgeleitete DE-Typen werden automatisch ergnzt.
		setDeTypes2Fg("typ.deLve", 1);
		setDeTypes2Fg("typ.deAxl", 2);
		setDeTypes2Fg("typ.deUfd", 3);
		setDeTypes2Fg("typ.deWzg", 4);
		setDeTypes2Fg("typ.deWww", 5);
		setDeTypes2Fg("typ.deVlt", 6);
		setDeTypes2Fg("typ.de128", 128);
		setDeTypes2Fg("typ.de253", 253);
		setDeTypes2Fg("typ.deSys", 254);
		
		/**
		 * Liste mit allen der Konfiguration bekannten Konfigurationsobjekte vom Typ typ.gert
		 */
		final List<SystemObject> devices = _configuration.getType("typ.gert").getElements();
		
		// Auszuwertende konfigurierende Attributgruppen
		final AttributeGroup atgGert = _configuration.getAttributeGroup("atg.gert");
		final AttributeGroup atgAnschlussPunktKommunikationsPartner = _configuration.getAttributeGroup("atg.anschlussPunktKommunikationsPartner");
		
		// Fr alle Gerte ein Tls-Knoten Objekt anlegen und in die Map einfgen 
		// Erster Lauf ber die Gerte, damit beim nchsten Lauf die zu refernzierenden Objekte bereits vorhanden sind.
		for(SystemObject device : devices) {
			final int nodeNumber = device.getConfigurationData(atgGert).getItem("KnotenNummer").asScaledValue().intValue();
			_tlsDevices.put(device, new TlsNode(device, nodeNumber));
		}
		
		// Ermittlung der Kinderknoten, des Vaterknotens und der vorhandenen FG (Bei Gerten vom Typ typ.steuerModul)
		// Schleife ber alle Gerte (Zeiter Lauf)
		for(SystemObject device : devices) {
			
			final List<SystemObject> connectionPointDevices = ((ConfigurationObject)device).getObjectSet("AnschlussPunkteGert").getElements();
			
			// Schleife ber alle Anschlusspunkte
			for(SystemObject connectionPointDevice : connectionPointDevices) {
				
				// Schleife ber alle Anschlusspunkte, die an dem Anschlusspunkt angeschlossen sind
				final List<SystemObject> connectionPointsPartner = ((ConfigurationObject)connectionPointDevice).getObjectSet(
				        "AnschlussPunkteKommunikationsPartner").getElements();
				
				for(SystemObject connectionPointPartner : connectionPointsPartner) {
					// Auswertung, der Referenz
					final Data dataAnschlussPunktKommunikationsPartner = connectionPointPartner.getConfigurationData(atgAnschlussPunktKommunikationsPartner);
					
					if(dataAnschlussPunktKommunikationsPartner != null) {
						final SystemObject partnerDevice = dataAnschlussPunktKommunikationsPartner
						        .getItem("KommunikationsPartner")
						        .asReferenceValue()
						        .getSystemObject();
						//						System.out.println("Fge dem Knoten " + device + " den Subknoten "+ partnerDevice + " hinzu");
						_tlsDevices.get(device).addSubNode(_tlsDevices.get(partnerDevice));
						_tlsDevices.get(partnerDevice).setParent(_tlsDevices.get(device));
					}
				}
			}
			
			// Bei Steuermodulen die DE auswerten
			if(device.getType().getPid().equals("typ.steuerModul")) {
				final List<SystemObject> eaks = ((ConfigurationObject)device).getObjectSet("Eak").getElements();
				for(SystemObject eak : eaks) {
					final List<SystemObject> des = ((ConfigurationObject)eak).getObjectSet("De").getElements();
					for(SystemObject de : des) {
						final String deTyp = de.getType().getPid();
						final Integer fg = _deTyp2FgNr.get(deTyp);
						if(fg == null) {
							_debug.warning("Keine Umsetzung fr DE vom Typ " + deTyp + " auf eine FG in TlsNode definiert");
						}
						else {
							_tlsDevices.get(device).addFg(fg);
						}
					}
				}
			}
		}
	}
	
	/**
	 * Funktion, die die Map mit der Zuordnung von DE-Typen zu Funktiongruppen aufbaut. Hierbei werden zu den bergebenen DE-Typen jeweils automatisch alle
	 * eventuell abgeleiteten DE-Typen ermittelt und ebenfalls in die Map aufgenommen.
	 * 
	 * @param deType
	 *            PID der DE-Typs
	 * @param fgNumber
	 *            Zugeordnete Funktionsgruppe
	 */
	private void setDeTypes2Fg(String deType, int fgNumber) {
		
		final SystemObjectType type = _configuration.getType(deType);
		
		_deTyp2FgNr.put(deType, fgNumber);
		
		final List<SystemObjectType> subTypes = type.getSubTypes();
		for(SystemObjectType subType : subTypes) {
			setDeTypes2Fg(subType.getPid(), fgNumber);
		}
	}
	
	/**
	 * Gibt den TLS-Knoten zu dem bergebenen Gert zurck.
	 * 
	 * @param systemObject
	 * @return TLS-Knoten
	 */
	public TlsNode getTlsNode(final SystemObject systemObject) {
		
		final TlsNode tlsNode = _tlsDevices.get(systemObject);
		if(tlsNode == null) throw new IllegalArgumentException("Das Systemobjekt muss vom Typ Gert sein");
		return tlsNode;
	}
	
}
