/*
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * Copyright 2007 by Kappich Systemberatung, Aachen
 * Copyright 2004 by Kappich+Kni Systemberatung Aachen (K2S)
 * 
 * This file is part of de.bsvrz.pat.sysbed.
 * 
 * de.bsvrz.pat.sysbed is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pat.sysbed is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pat.sysbed; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.pat.sysbed.dataview;

import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.sys.funclib.debug.Debug;

import java.util.*;

/**
 * Diese Klasse erhlt von der Applikation Datenstze und Informationen. Hierbei handelt es sich um zustzliche Datenstze, um genderte Datenstze, um eine
 * Menge neuer Datenstze oder um Informationen darber, welche Datenstze gelscht wurden. Diese Datenstze und Informationen werden empfangen, ggf.
 * verarbeitet und an alle angemeldeten {@link DataViewPanel DataViewPanels} bergeben.
 *
 * @author Kappich Systemberatung
 * @version $Revision: 6361 $
 */
public class DataViewModel {

	/** der Debug-Logger */
	private final Debug _debug = Debug.getLogger();

	/** speichert alle angemeldeten Listener */
	private final List<DataViewListener> _listener;

	/** speichert alle Datenstze, die von der Applikation bergeben werden */
	private final List<DataTableObject> _datasets;

	/** speichert die darzustellende Attributgruppe */
	private final AttributeGroup _attributeGroup;

	/** speichert zu jedem Objekt den letzten zu betrachtenden Datensatz */
	private final Map<Long, DataTableObject> _actualDatasets = new HashMap<Long, DataTableObject>();

	/* ################ Konstruktor ############## */
	/**
	 * Konstruktor.
	 *
	 * @param attributeGroup Attributgruppe, die dargestellt werden soll
	 */
	public DataViewModel(final AttributeGroup attributeGroup) {
		_attributeGroup = attributeGroup;
		_listener = new LinkedList<DataViewListener>();
		_datasets = new ArrayList<DataTableObject>();
	}


	/* ################# Methoden fr das DataViewPanel ########### */
	/**
	 * Gibt die Attributgruppe zurck.
	 *
	 * @return Attributgruppe
	 */
	public AttributeGroup getAttributeGroup() {
		return _attributeGroup;
	}

	/**
	 * Fgt einen Datensatz an bestehende Daten hinten an.
	 *
	 * @param dataTableObject neuer Datensatz
	 */
	public void addDatasetBelow(final DataTableObject dataTableObject) {
		final List<DataTableObject> dataTableObjects = new ArrayList<DataTableObject>(1);
		dataTableObjects.add(dataTableObject);
		addDatasetsBelow(dataTableObjects);
	}

	/**
	 * Fgt mehrere Datenstze an bestehende Daten hinten an.
	 *
	 * @param dataTableObjects Liste mit anzuzeigenden Datenstzen
	 */
	public void addDatasetsBelow(final List<DataTableObject> dataTableObjects) {
		_datasets.addAll(dataTableObjects);	// einmal werden die neuen Datenstze hier abgespeichert
		fireAddDatasets(dataTableObjects); // zum anderen werden die neuen Datenstze weitergereicht zur Online-Tabelle
	}

	/**
	 * Fgt einen neuen Datensatz oben in der Tabelle ein.
	 *
	 * @param dataTableObject der neue Datensatz
	 */
	public void addDatasetAbove(final DataTableObject dataTableObject) {
		_datasets.add(0, dataTableObject);
		fireAddDataset(0, dataTableObject);
	}

	/**
	 * bergibt initial alle in Frage kommenden Datenstze. Bestehende werden gelscht. Durch unten genannte Methoden knnen weitere Datenstze hinzugefgt
	 * werden.
	 *
	 * @param dataTableObjects Liste mit den Datenstzen
	 *
	 * @see #addDatasetAbove(DataTableObject)
	 * @see #addDatasetBelow(DataTableObject)
	 */
	public void setDatasets(final List<DataTableObject> dataTableObjects) {
		_datasets.clear();
		_datasets.addAll(dataTableObjects);
		fireSetDatasets(dataTableObjects);
	}

	/**
	 * Stellt zu den angemeldeten Objekten jeweils den aktuellsten Datensatz dar. Kommen neue Daten vom Datenverteiler, werden auch nur die Daten der Objekte
	 * ersetzt, die neu sind. Alle anderen Daten bleiben erhalten.
	 *
	 * @param dataTableObjects auszuwertende Daten
	 */
	public void updateDatasets(final List<DataTableObject> dataTableObjects) {
		// zuerst alles was kommt in die Map einspeisen
		for(DataTableObject dataTableObject : dataTableObjects) {
			_actualDatasets.put(dataTableObject.getObject().getId(), dataTableObject);
		}
		// anschlieend zu allen Systemobjekten ihren Wert (=letzter gltiger Wert) rausholen
		final List<DataTableObject> datasetList = new LinkedList<DataTableObject>();
		Set<Long> keySet = _actualDatasets.keySet();
		for(Long objectID : keySet) {
			datasetList.add(_actualDatasets.get(objectID));
		}
		fireSetDatasets(datasetList);
	}

//	/**
//	 * Sucht nach dem ersten Vorkommen dieses Datensatzes und lscht ihn.
//	 *
//	 * @param dataTableObject zu lschenden Datensatz
//	 *
//	 * @return gibt an, ob der Datensatz gelscht wurde
//	 */
//	public boolean removeDataset(final DataTableObject dataTableObject) {
//		int index = _datasets.indexOf(dataTableObject);
//		if(index == -1) {
//			_debug.warning("Zu lschender Datensatz ist nicht vorhanden!");
//			return false;
//		}
//		else {
//			_datasets.remove(index);
//			fireRemoveDataset(index);
//			return true;
//		}
//	}


	/* ################ Listener - Funktionalitt ############### */
	/**
	 * Meldet einen Listener beim Model an.
	 *
	 * @param listener der anzumeldende Listener
	 */
	public void addDataViewListener(final DataViewListener listener) {
		_listener.add(listener);
	}

	/**
	 * Meldet einen Listener vom Model ab.
	 *
	 * @param listener der abzumeldende Listener
	 */
	public void removeDataViewListener(final DataViewListener listener) {
		_listener.remove(listener);
	}

	/**
	 * Alle Listener werden benachrichtigt, dass einer oder mehrere Datenstze an die bestehenden Datenstze angefgt wurden. Diese werden auch bergeben.
	 *
	 * @param datasets neue Datenstze
	 */
	private void fireAddDatasets(final List<DataTableObject> datasets) {
		List<DataTableObject> unmodifiableList = Collections.unmodifiableList(datasets);
		for(Iterator iterator = _listener.iterator(); iterator.hasNext();) {
			DataViewListener dataViewListener = (DataViewListener)iterator.next();
			dataViewListener.addDatasets(unmodifiableList);
		}
	}

	/**
	 * Alle Listener werden benachrichtigt, dass ein Datensatz an eine bestimmte Position der bestehenden Datenstze eingefgt wurde.
	 *
	 * @param index           Position des neuen Datensatzes
	 * @param dataTableObject der neue Datensatz
	 */
	private void fireAddDataset(int index, final DataTableObject dataTableObject) {
		for(Iterator iterator = _listener.iterator(); iterator.hasNext();) {
			DataViewListener dataViewListener = (DataViewListener)iterator.next();
			dataViewListener.addDataset(index, dataTableObject);
		}
	}

	/**
	 * Alle Listener werden benachrichtigt, dass die bisherigen Datenstze gelscht und durch die neuen (bergebenen) Datenstze ersetzt werden. Diese werden auch
	 * bergeben.
	 *
	 * @param datasets neue Datenstze
	 */
	private void fireSetDatasets(final List<DataTableObject> datasets) {
		final List<DataTableObject> unmodifiableList = Collections.unmodifiableList(datasets);
		for(DataViewListener dataViewListener : _listener) {
			dataViewListener.setDatasets(unmodifiableList);
		}
	}

	/**
	 * Alle Listener werden benachrichtigt, dass der Datensatz an der Position <code>index</code> gelscht wurde.
	 *
	 * @param index Position des gelschten Datensatzes
	 */
	private void fireRemoveDataset(int index) {
		for(Iterator iterator = _listener.iterator(); iterator.hasNext();) {
			DataViewListener dataViewListener = (DataViewListener)iterator.next();
			dataViewListener.removeDataset(index);
		}
	}
}
