/*
 * Copyright 2007 by Kappich Systemberatung, Aachen
 * Copyright 2006 by Kappich Systemberatung, Aachen
 * Copyright 2004 by Kappich+Kni Systemberatung Aachen (K2S)
 * 
 * This file is part of de.bsvrz.pat.sysbed.
 * 
 * de.bsvrz.pat.sysbed is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pat.sysbed is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pat.sysbed; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.pat.sysbed.dataview;

import de.bsvrz.dav.daf.main.config.Attribute;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.AttributeListDefinition;
import de.bsvrz.dav.daf.main.config.AttributeType;
import de.bsvrz.sys.funclib.debug.Debug;

import javax.swing.*;
import javax.swing.border.EtchedBorder;
import java.awt.*;
import java.awt.event.AdjustmentEvent;
import java.awt.event.AdjustmentListener;
import java.util.*;
import java.util.List;

/**
 * Dieses Panel stellt die Datenstze, die das {@link DataViewModel Model} bergibt, in einem ScrollPane dar. Mit dem Konstruktor wird ein baumartiger Header
 * erstellt, dessen Spaltenbreiten durch Schieberegler verndert werden kann. Damit dies funktioniert, muss nachdem der Header erzeugt und dieser angezeigt
 * (z.B. durch validate() oder durch Frame.setVisible()) wird mit der Methode {@link #initHeaderSize()} initialisiert werden.
 *
 * @author Kappich Systemberatung
 * @version $Revision: 5052 $
 */
public class DataViewPanel extends JPanel implements DataViewListener {

	/** Der Debug-Logger */
	private final Debug _debug = Debug.getLogger();

	/* ############## Variablen ############## */
	/** speichert eine Instanz des DataViewModel */
	private final DataViewModel _dataViewModel;

	/** speichert die Attributgruppe */
	private final AttributeGroup _attributeGroup;

	/** speichert alle Datenstze, die von der Applikation gesendet wurden */
	private final List<Dataset> _datasets = new ArrayList<Dataset>();

	/** speichert das dargestellte ScrollPane */
	private final JScrollPane _headerScrollPane = new JScrollPane();

	/** speichert die Sruktur des Headers */
	private HeaderGrid _head;

	/** merkt sich, ob es sich um den ersten Aufruf handelt */
	private boolean _firstRun = true;

	/** speichert den vertikalen Scrollbalken */
	private final JScrollBar _verticalScrollBar = _headerScrollPane.getVerticalScrollBar();

	/** der obere Platzhalter fr den Zeilenheader */
	private final JPanel _upperRowHeaderPanel = new JPanel();

	/** der obere Platzhalter fr den Viewport */
	private final JPanel _upperViewportPanel = new JPanel();

	/** der sichtbare Teil des Zeilenheaders */
	private final JPanel _rowHeaderPanel = new JPanel();

	/** der sichtbare Teil des Viewports */
	private final JPanel _viewportPanel = new JPanel();

	/** der untere Platzhalter des Zeilenheaders */
	private final JPanel _lowerRowHeaderPanel = new JPanel();

	/** der untere Platzhalter des Viewports */
	private final JPanel _lowerViewportPanel = new JPanel();

	/** Gibt den ersten Datensatz an, der angezeigt wird. */
	private int _firstRow = 0;

	/** Gibt den letzten Datensatz an, der angezeigt wird. */
	private int _lastRow = 0;

	/** die sichtbare Hhe des Viewports */
	private int _screenHeight;

	/** speichert das Zeitformat fr die Spalte 'Zeit' */
	private final String TIME_FORMAT = "dd.MM.yyyy HH:mm:ss,SSS";

	/** Gibt die Breite des Zeilenheaders zurck. */
	private int _rowHeaderWidth = 0;

	/* ########### Konstruktor und Aufbau des Panels ########### */

	/**
	 * Konstruktor. Bentigt eine Instanz des {@link DataViewModel}. Das Model gibt Benachrichtigungen, falls sich an den anzuzeigenden Daten etwas ndert.
	 *
	 * @param model das DataViewModel
	 */
	public DataViewPanel(DataViewModel model) {
		_dataViewModel = model;
		_attributeGroup = _dataViewModel.getAttributeGroup();
		createAndShowGui();
	}

	/** Erstellt die Elemente fr das Panel und ordnet diese an. */
	private void createAndShowGui() {
		setScrollPane();

		// anordnen des ScrollPanes auf dem Panel (DataViewPanel)
		setLayout(new BorderLayout());
		add(_headerScrollPane, BorderLayout.CENTER);
	}

	/**
	 * Das ScrollPane wird mit Spaltenheader, Zeilenheader und oberer linker Ecke ausgestattet und initialisiert. Die vertikale Scrollbar wird ebenfalls
	 * initialisiert und ein {@link VerticalScrollBarAdjustmentListener AdjustmentListener} wird angemeldet.
	 */
	private void setScrollPane() {
		// Initialisierung des oberen und unteren Platzhalters
		setUpperPanel(0);
		setLowerPanel(0);

		// Aufbau des ScrollPanes
		_headerScrollPane.setColumnHeaderView(createColumnHeader());
		_headerScrollPane.setRowHeaderView(createRowHeader());
		_headerScrollPane.setCorner(JScrollPane.UPPER_LEFT_CORNER, createUpperLeftCorner());
		_headerScrollPane.setViewportView(createViewport());

		// Aufbau des vertikalen ScrollBars
		_headerScrollPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
		_verticalScrollBar.setUnitIncrement(20);	// eine Zeile hat momentan 20 Pixel in der Hhe!
		_verticalScrollBar.setBlockIncrement(20);   // wie oben
		_verticalScrollBar.addAdjustmentListener(new VerticalScrollBarAdjustmentListener());

		// Horizontale Scrollbar
		JScrollBar horizontalScrollBar = _headerScrollPane.getHorizontalScrollBar();
		horizontalScrollBar.setUnitIncrement(30);
		horizontalScrollBar.setBlockIncrement(30);
	}

	/**
	 * Erstellt den vollstndigen Spaltenheader fr das ScrollPane.
	 *
	 * @return Spaltenheader
	 */
	private JPanel createColumnHeader() {
		_head = new HeaderGrid(null, _attributeGroup, this);
		setHeaderSuccessors(_head);
		JPanel westHeader = new JPanel(new BorderLayout());
		westHeader.add(_head.createHeader(), BorderLayout.WEST);
		return westHeader;
	}

	/**
	 * Erstellt den initialen Zeilenheader fr das ScrollPane.
	 *
	 * @return Zeilenheader
	 */
	private JPanel createRowHeader() {
		JPanel rowHeader = new JPanel();
		rowHeader.setLayout(new BoxLayout(rowHeader, BoxLayout.Y_AXIS));
		_rowHeaderPanel.setLayout(new BoxLayout(_rowHeaderPanel, BoxLayout.Y_AXIS));
		rowHeader.add(_upperRowHeaderPanel);
		rowHeader.add(_rowHeaderPanel);
		rowHeader.add(_lowerRowHeaderPanel);

		JPanel westRowHeader = new JPanel(new BorderLayout());
		westRowHeader.add(rowHeader, BorderLayout.WEST);

		return westRowHeader;
	}

	/**
	 * Erstellt die Komponente fr die obere linke Ecke im ScrollPane. Sie zeigt die Zeit und das Objekt an.
	 *
	 * @return die Komponente
	 */
	private JComponent createUpperLeftCorner() {
		JLabel dataKindLabel = new JLabel("Art");
		dataKindLabel.setBorder(new EtchedBorder());
		dataKindLabel.setHorizontalAlignment(SwingConstants.CENTER);
		JLabel timeLabel = new JLabel("Zeit");
		timeLabel.setBorder(new EtchedBorder());
		timeLabel.setHorizontalAlignment(SwingConstants.CENTER);
		JLabel objectLabel = new JLabel("Objekt");
		objectLabel.setBorder(new EtchedBorder());
		objectLabel.setHorizontalAlignment(SwingConstants.CENTER);

		// passt die Breite dieser Headerspalte mit der Breite bei den Datenstzen?
		JPanel dataKindPanel = new JPanel(new BorderLayout());
		dataKindPanel.add(dataKindLabel, BorderLayout.CENTER);

		JPanel gridPanel = new JPanel(new GridLayout(1, 2));
		gridPanel.add(timeLabel);
		gridPanel.add(objectLabel);

		JPanel panel = new JPanel(new BorderLayout());
		panel.add(dataKindPanel, BorderLayout.WEST);
		panel.add(gridPanel, BorderLayout.CENTER);

		return panel;
	}

	/**
	 * Erstellt den initialen Viewport fr das Scrollpane
	 *
	 * @return Viewport des Scrollpanes
	 */
	private JPanel createViewport() {
		JPanel viewport = new JPanel();
		viewport.setLayout(new BoxLayout(viewport, BoxLayout.Y_AXIS));
		_viewportPanel.setLayout(new BoxLayout(_viewportPanel, BoxLayout.Y_AXIS));
		viewport.add(_upperViewportPanel);
		viewport.add(_viewportPanel);
		viewport.add(_lowerViewportPanel);

		JPanel westViewport = new JPanel(new BorderLayout());
		westViewport.add(viewport, BorderLayout.WEST);

		return westViewport;
	}

	/* ################ Hilfsmethoden fr das DataViewPanel ################### */

	/**
	 * Gibt den Spaltenheader zurck.
	 *
	 * @return den Spaltenheader
	 */
	public HeaderGrid getHeaderGrid() {
		return _head;
	}


	/**
	 * Die Gre des Headers wird initial gesetzt. Dies ist notwendig, damit die Headergre durch die Maus verndert werden kann. Diese Methode muss nach dem
	 * Anzeigen der Tabelle, allerdings bevor die ersten Daten kommen, aufgerufen werden.
	 *
	 * @see #setHeaderSizes(HeaderGrid,int)
	 */
	public void initHeaderSize() {
		HeaderGrid.HeaderElement element = _head.getHeaderElement();
		int elementWidth = element.getSize().width;
		int viewportWidth = _headerScrollPane.getHorizontalScrollBar().getVisibleAmount() - _rowHeaderWidth;	// 350 fr den RowHeader
		int percent;
		if(elementWidth == 0 || elementWidth >= viewportWidth) {
			percent = 0;
		}
		else {
			if(elementWidth >= 500) {
				percent = 100;
			}
			else {
				percent = (int)(500 * 100) / elementWidth;
			}
		}
		setHeaderSizes(_head, percent);
		_headerScrollPane.validate();
	}

	/**
	 * Initialisiert die Gren aller Elemente des Headers rekursiv.
	 *
	 * @param headerGrid Spaltenheader
	 * @param percent    gibt die prozentuale Ausdehnung an
	 */
	private void setHeaderSizes(HeaderGrid headerGrid, int percent) {
		HeaderGrid.HeaderElement element = headerGrid.getHeaderElement();
		element.setMinimumSize(new Dimension(0, 0));	// wird vom GridBagLayout nicht bentigt
		element.setMaximumSize(new Dimension(0, 0));	// wie setMinimumSize()
		Dimension size = element.getSize();
		headerGrid.getSplitter().setPreferredSize(new Dimension(5, size.height));
		if(headerGrid.getHeaderSuccessors().size() > 0) {
			element.setPreferredSize(new Dimension(0, size.height));
			int numberOfColumns = 0;
			for(Iterator iterator = headerGrid.getHeaderSuccessors().iterator(); iterator.hasNext();) {
				HeaderGrid successor = (HeaderGrid)iterator.next();
				setHeaderSizes(successor, percent);
				numberOfColumns += successor.getNumberOfColumns();
			}
			headerGrid.setNumberOfColumns(numberOfColumns);
		}
		else {	// Blattknoten
			headerGrid.setNumberOfColumns(1);
			
			int newWidth;
			if(percent == 0) {
				newWidth = size.width + 1;
			}
			else {
				newWidth = (int)((size.width * percent) / 100);
			}
			Dimension dim = new Dimension(newWidth, size.height);
			element.setPreferredSize(dim);
			element.setSize(dim);
		}
	}

	/**
	 * Bestimmt die Nachfolger einer Attributgruppe bzw. einer Attributliste.
	 *
	 * @param attributeList Objekt vom Typ <code>HeaderGrid</code>, von dem die Nachfolger bestimmt werden sollen.
	 */
	private void setHeaderSuccessors(HeaderGrid attributeList) {
		/* Nachfolger bestimmen und anfgen */
		Object node = attributeList.getHeaderElement().getObject();
		if(node instanceof AttributeGroup) {   // es handelt sich um die Attributgruppe
			AttributeGroup atg = (AttributeGroup)node;
			List attributes = atg.getAttributes();
			for(Iterator iterator = attributes.iterator(); iterator.hasNext();) {
				Attribute attribute = (Attribute)iterator.next();
				HeaderGrid successor = new HeaderGrid(attributeList, attribute, this);
				attributeList.addHeaderSuccessor(successor);
				setHeaderSuccessors(successor);   // weitere Nachfolger?
			}
		}
		else if(node instanceof Attribute) { // es handelt sich um ein Nachfolgeelement
			Attribute attribute = (Attribute)node;
			AttributeType attributeType = attribute.getAttributeType();
			if(attributeType instanceof AttributeListDefinition) { // Nachfolgeelement ist Attributliste
				AttributeListDefinition attributeListDefinition = (AttributeListDefinition)attributeType;
				List attributes = attributeListDefinition.getAttributes();
				for(Iterator iterator = attributes.iterator(); iterator.hasNext();) {
					Attribute att = (Attribute)iterator.next();
					HeaderGrid successor = new HeaderGrid(attributeList, att, this);
					attributeList.addHeaderSuccessor(successor);
					setHeaderSuccessors(successor);
				}
			}
		}
	}

	/**
	 * Hoffentlich bald berflssig. Sptestens, wenn die Methode {@link #initHeaderSize()} von der Applikation aufgerufen wird.
	 *
	 * @return gibt true oder false zurck
	 */
	public boolean getFirstRun() {
		return _firstRun;
	}

	/**
	 * Kann auf false gesetzt werden, falls die Gren des Spaltenheaders initialisiert worden sind.
	 *
	 * @param firstRun true ist Default-Einstellung
	 */
	public void setFirstRun(boolean firstRun) {
		_firstRun = firstRun;
	}

	/* ################### Methoden des DataViewListeners ############ */

	/**
	 * Fgt eine beliebige Anzahl neuer Datenstze an die bestehenden Datenstze hinten an.
	 *
	 * @param dataTableObjects Liste neuer Datenstze
	 */
	public void addDatasets(final List<DataTableObject> dataTableObjects) {
		if(getFirstRun()) {
			initHeaderSize();
			setFirstRun(false);
		}
		SwingUtilities.invokeLater(
				new Runnable() {
					public void run() {
						for(DataTableObject dataTableObject : dataTableObjects) {
							Dataset dataset = new Dataset(_head, dataTableObject);
							_datasets.add(dataset);

							boolean check = checkRowHeight();
							if(check) {
								_lastRow = _datasets.indexOf(dataset);

								dataset.setLinks();
								JComponent rowHeaderRow = dataset.getRowHeaderRow(TIME_FORMAT);
								JComponent viewportRow = dataset.getViewportRow();

								int height = dataset.getHeight();
								rowHeaderRow.setPreferredSize(new Dimension(rowHeaderRow.getPreferredSize().width, height));
								rowHeaderRow.setMaximumSize(new Dimension(rowHeaderRow.getMaximumSize().width, height));
								viewportRow.setMaximumSize(new Dimension(viewportRow.getMaximumSize().width, height));

								_rowHeaderPanel.add(rowHeaderRow);
								_viewportPanel.add(viewportRow);
								_headerScrollPane.validate();
							}
							else {
								// Hhe = alte Hhe des Panels + neue Hhe
								int lowerHeight = _lowerViewportPanel.getSize().height;
								lowerHeight += dataset.getHeight();
								setLowerPanel(lowerHeight);
								_headerScrollPane.revalidate();
							}
						}
					}
				}
		);
	}

	/**
	 * Fgt einen Datensatz an eine bestimmte Position der bisherigen Datenstze ein. Befindet sich die Position innerhalb der gerade dargestellten Datenstze,
	 * dann wird die Darstellung neu erzeugt und angezeigt.
	 *
	 * @param index           Position des neuen Datensatzes
	 * @param dataTableObject der neue Datensatz
	 */
	public void addDataset(final int index, final DataTableObject dataTableObject) {
		if(getFirstRun()) {
			initHeaderSize();
			setFirstRun(false);
		}
		SwingUtilities.invokeLater(
				new Runnable() {
					public void run() {
						Dataset dataset = new Dataset(_head, dataTableObject);
						int upperHeight = _upperViewportPanel.getSize().height;
						int viewportHeight = _viewportPanel.getSize().height;
						int lowerHeight = _lowerViewportPanel.getSize().height;
						int value = _verticalScrollBar.getModel().getValue();
						_screenHeight = _verticalScrollBar.getVisibleAmount();

						if(_datasets.isEmpty()) {
							// einfach hinzufgen, _firstRow, _lastRow setzen
							_datasets.add(dataset);
							_firstRow = 0;
							_lastRow = 0;

							dataset.setLinks();
							JComponent rowHeaderRow = dataset.getRowHeaderRow(TIME_FORMAT);
							JComponent viewportRow = dataset.getViewportRow();

							int nextRowHeight = dataset.getHeight();

							rowHeaderRow.setPreferredSize(new Dimension(rowHeaderRow.getPreferredSize().width, nextRowHeight));
							rowHeaderRow.setMaximumSize(new Dimension(rowHeaderRow.getMaximumSize().width, nextRowHeight));
							viewportRow.setMaximumSize(new Dimension(viewportRow.getMaximumSize().width, nextRowHeight));

							_rowHeaderPanel.add(rowHeaderRow);
							_viewportPanel.add(viewportRow);
						}
						else {
							// an Position index einen Datensatz einfgen
							if(index < _firstRow) {
								int height = dataset.getHeight();
								setUpperPanel(_upperViewportPanel.getSize().height + height);
								_datasets.add(index, dataset);
								_firstRow++;
								_lastRow++;
								_headerScrollPane.validate();
								_verticalScrollBar.getModel().setValue(_verticalScrollBar.getModel().getValue() + height);
							}
							else if(index > _lastRow) {
								int height = dataset.getHeight();
								setLowerPanel(_lowerViewportPanel.getSize().height + height);
								_datasets.add(index, dataset);
							}
							else {	// index >= _firstRow && index <= _lastRow
								// erst einfgen, dann ggf. lschen
								int position = index - _firstRow;

								dataset.setLinks();

								JComponent rowHeaderRow = dataset.getRowHeaderRow(TIME_FORMAT);
								JComponent viewportRow = dataset.getViewportRow();

								int nextRowHeight = dataset.getHeight();
								viewportHeight += nextRowHeight;

								rowHeaderRow.setPreferredSize(new Dimension(rowHeaderRow.getPreferredSize().width, nextRowHeight));
								rowHeaderRow.setMaximumSize(new Dimension(rowHeaderRow.getMaximumSize().width, nextRowHeight));
								viewportRow.setMaximumSize(new Dimension(viewportRow.getMaximumSize().width, nextRowHeight));

								_rowHeaderPanel.add(rowHeaderRow, position);
								_viewportPanel.add(viewportRow, position);
								_datasets.add(index, dataset);
								_lastRow++;

								// neue letzte zeile ermitteln
								Dataset firstDataset = (Dataset)_datasets.get(_firstRow);
								int firstRowHeight = firstDataset.getHeight();
								viewportHeight = 0;
								int newLastRow = _firstRow;
								while(upperHeight + viewportHeight < value + _screenHeight + firstRowHeight && newLastRow < _datasets.size()) {
									Dataset nextDataset = (Dataset)_datasets.get(newLastRow++);
									viewportHeight += nextDataset.getHeight();
								}
								newLastRow--;
								for(int i = _lastRow; i > newLastRow; i--) {   // berflssige Zeilen lschen
									Dataset nextDataset = (Dataset)_datasets.get(i);
									lowerHeight += nextDataset.getHeight();
									_rowHeaderPanel.remove(_rowHeaderPanel.getComponentCount() - 1);
									_viewportPanel.remove(_viewportPanel.getComponentCount() - 1);
									nextDataset.unsetLinks();
								}
								_lastRow = newLastRow;
								setLowerPanel(lowerHeight);
							}
						}
						_headerScrollPane.revalidate();
					}
				}
		);
	}

	/**
	 * Alle bisherigen Datenstze werden gelscht und die neuen werden bernommen. Die ersten Datenstze, die angezeigt werden knnen, werden dargestellt.
	 *
	 * @param dataTableObjects Liste neuer Datenstze
	 */
	public void setDatasets(final List<DataTableObject> dataTableObjects) {
		if(getFirstRun()) {
			initHeaderSize();
			setFirstRun(false);
		}
		SwingUtilities.invokeLater(
				new Runnable() {
					public void run() {
						// alle bisherigen Datenstze lschen
						
						for(Dataset dataset : _datasets) {
							dataset.unsetLinks();
						}
						_datasets.clear();
						_firstRow = 0;
						_lastRow = -1;

						// die Anzeige initialisieren
						setUpperPanel(0);
						while(_viewportPanel.getComponentCount() > 0) {
							_rowHeaderPanel.remove(0);
							_viewportPanel.remove(0);
						}
						_rowHeaderPanel.setSize(new Dimension(_rowHeaderPanel.getSize().width, 0));
						_viewportPanel.setSize(new Dimension(_viewportPanel.getSize().width, 0));
						setLowerPanel(0);
						for(DataTableObject dataTableObject : dataTableObjects) {
							Dataset dataset = new Dataset(_head, dataTableObject);
							_datasets.add(dataset);

							boolean check = checkRowHeight();
							if(check) {
								_lastRow = _datasets.indexOf(dataset);

								dataset.setLinks();
								JComponent rowHeaderRow = dataset.getRowHeaderRow(TIME_FORMAT);
								JComponent viewportRow = dataset.getViewportRow();

								int height = dataset.getHeight();
								rowHeaderRow.setPreferredSize(new Dimension(rowHeaderRow.getPreferredSize().width, height));
								rowHeaderRow.setMaximumSize(new Dimension(rowHeaderRow.getMaximumSize().width, height));
								viewportRow.setMaximumSize(new Dimension(viewportRow.getMaximumSize().width, height));

								_rowHeaderPanel.add(rowHeaderRow);
								_viewportPanel.add(viewportRow);
								_headerScrollPane.validate();
							}
							else {
								// Hhe = alte Hhe des Panels + neue Hhe
								int lowerHeight = _lowerViewportPanel.getSize().height;
								lowerHeight += dataset.getHeight();
								setLowerPanel(lowerHeight);
								_headerScrollPane.revalidate();
							}
						}
					}
				}
		);
	}

	/**
	 * Lscht einen Datensatz an angegebener Position.
	 *
	 * @param index Position des zu lschenden Datensatzes
	 */
	public void removeDataset(final int index) {
		SwingUtilities.invokeLater(
				new Runnable() {
					public void run() {
						// gibt es diesen Index berhaupt?
						int upperHeight = _upperViewportPanel.getSize().height;
						int viewportHeight = _viewportPanel.getSize().height;
						int lowerHeight = _lowerViewportPanel.getSize().height;
						int value = _verticalScrollBar.getModel().getValue();
						_screenHeight = _verticalScrollBar.getVisibleAmount();

//				System.out.println("First: " + _firstRow + " \t Last: " + _lastRow);
						if(index >= _firstRow && index <= _lastRow) {
							// welche Zeile lschen

							Dataset dataset = (Dataset)_datasets.get(index);
							viewportHeight -= dataset.getHeight();

							dataset.unsetLinks();
							_datasets.remove(index);
							int row = index - _firstRow;
							_rowHeaderPanel.remove(row);
							_viewportPanel.remove(row);
							// ggf. neue hinzufgen - aber nur, wenn Datensatz existiert
							if(_datasets.size() > 0 && _firstRow < _datasets.size()) {
								Dataset firstDataset = (Dataset)_datasets.get(_firstRow);
								int firstRowHeight = firstDataset.getHeight();
								while((upperHeight + viewportHeight < value + _screenHeight + firstRowHeight) && (_lastRow < _datasets.size())) {
									Dataset lastDataset = (Dataset)_datasets.get(_lastRow++);
									lastDataset.setLinks();

									JComponent rowHeaderRow = lastDataset.getRowHeaderRow(TIME_FORMAT);
									JComponent viewportRow = lastDataset.getViewportRow();

									int lastRowHeight = lastDataset.getHeight();
									viewportHeight += lastRowHeight;
									lowerHeight -= lastRowHeight;

									rowHeaderRow.setPreferredSize(new Dimension(rowHeaderRow.getPreferredSize().width, lastRowHeight));
									rowHeaderRow.setMaximumSize(new Dimension(rowHeaderRow.getMaximumSize().width, lastRowHeight));
									viewportRow.setMaximumSize(new Dimension(viewportRow.getMaximumSize().width, lastRowHeight));

									_rowHeaderPanel.add(rowHeaderRow);
									_viewportPanel.add(viewportRow);
								}
								_lastRow--;

								int anzahl = _viewportPanel.getComponentCount();
								if(anzahl != ((_lastRow - _firstRow) + 1)) {
									_debug.error("Anzahl der Komponenten stimmt nicht mit _firstRow und _lastRow berein!");
								}
							}
						}
						else {
							Dataset dataset = (Dataset)_datasets.get(index);
							int height = dataset.getHeight();
							if(index < _firstRow) {
								upperHeight -= height;
								_verticalScrollBar.getModel().setValue(_verticalScrollBar.getModel().getValue() - height);
							}
							else {	// index > _lastRow
								lowerHeight -= height;
							}
							_datasets.remove(index);	// Datensatz lschen
						}
						setUpperPanel(upperHeight);
						setLowerPanel(lowerHeight);
						_headerScrollPane.validate();
					}
				}
		);
	}

	/**
	 * Aktualisiert an angegebener Position den Datensatz.
	 *
	 * @param index           Position des zu aktualisierenden Datensatzes
	 * @param dataTableObject aktueller Datensatz
	 */
	public void update(final int index, final DataTableObject dataTableObject) {
		SwingUtilities.invokeLater(
				new Runnable() {
					public void run() {
						if(index >= _firstRow && index <= _lastRow) {
							int upperHeight = _upperViewportPanel.getSize().height;
							int viewportHeight = _viewportPanel.getSize().height;
							int lowerHeight = _lowerViewportPanel.getSize().height;
							int value = _verticalScrollBar.getModel().getValue();
							_screenHeight = _verticalScrollBar.getVisibleAmount();

							int position = index - _firstRow;

							// entfernen
							_rowHeaderPanel.remove(position);
							_viewportPanel.remove(position);

							Dataset dataset = (Dataset)_datasets.get(index);
							int height = dataset.getHeight();
							viewportHeight -= height;
							dataset.unsetLinks();
							// einfgen
							Dataset newDataset = new Dataset(_head, dataTableObject);
							_datasets.set(index, newDataset);
							newDataset.setLinks();
							height = newDataset.getHeight();
							viewportHeight += height;

							JComponent rowHeaderRow = newDataset.getRowHeaderRow(TIME_FORMAT);
							JComponent viewportRow = newDataset.getViewportRow();

							rowHeaderRow.setPreferredSize(new Dimension(rowHeaderRow.getPreferredSize().width, height));
							rowHeaderRow.setMaximumSize(new Dimension(rowHeaderRow.getMaximumSize().width, height));
							viewportRow.setMaximumSize(new Dimension(viewportRow.getMaximumSize().width, height));

							_rowHeaderPanel.add(rowHeaderRow, position);
							_viewportPanel.add(viewportRow, position);

							// ist die Hhe auch hoch genug? ist auch genug im Fenster sichtbar?
							Dataset firstDataset = (Dataset)_datasets.get(_firstRow);
							int firstRowHeight = firstDataset.getHeight();
							if(upperHeight + viewportHeight < value + _screenHeight + firstRowHeight) {
								_lastRow++;
								while((upperHeight + viewportHeight < value + _screenHeight + firstRowHeight) && (_lastRow < _datasets.size())) {
									Dataset lastDataset = (Dataset)_datasets.get(_lastRow++);
									lastDataset.setLinks();

									JComponent rowHeaderRow2 = lastDataset.getRowHeaderRow(TIME_FORMAT);
									JComponent viewportRow2 = lastDataset.getViewportRow();

									int lastRowHeight = lastDataset.getHeight();
									viewportHeight += lastRowHeight;
									lowerHeight -= lastRowHeight;

									rowHeaderRow2.setPreferredSize(new Dimension(rowHeaderRow2.getPreferredSize().width, lastRowHeight));
									rowHeaderRow2.setMaximumSize(new Dimension(rowHeaderRow2.getMaximumSize().width, lastRowHeight));
									viewportRow2.setMaximumSize(new Dimension(viewportRow2.getMaximumSize().width, lastRowHeight));

									_rowHeaderPanel.add(rowHeaderRow2);
									_viewportPanel.add(viewportRow2);
								}
								_lastRow--;

								// untere Platzhalter bestimmen
								setLowerPanel(lowerHeight);
							}
							int anzahl = _viewportPanel.getComponentCount();
							if(anzahl != ((_lastRow - _firstRow) + 1)) {
								_debug.error("Anzahl der Komponenten stimmt nicht mit _firstRow und _lastRow berein!");
							}
							_headerScrollPane.validate();
						}
						else {
							Dataset dataset = new Dataset(_head, dataTableObject);
							Dataset oldDataset = (Dataset)_datasets.get(index);
							int diff = dataset.getHeight() - oldDataset.getHeight();
							_datasets.set(index, dataset);
							if(index < _firstRow) {
								setUpperPanel(_upperViewportPanel.getSize().height + diff);
							}
							else {	// index > _lastRow
								setLowerPanel(_lowerViewportPanel.getSize().height + diff);
							}
							_headerScrollPane.validate();
						}
					}
				}
		);
	}

/* ############## Methoden fr das ScrollPane ############ */

	/**
	 * Setzt fr die oberen Platzhalter des Zeilenheaders und des Viewports die Hhe.
	 *
	 * @param height die neue Hhe der Platzhalter
	 */
	private void setUpperPanel(final int height) {
		Dimension size = new Dimension(1, height);
		_upperRowHeaderPanel.setPreferredSize(size);
		_upperRowHeaderPanel.setSize(size);
		_upperRowHeaderPanel.setMinimumSize(size);
		_upperRowHeaderPanel.setMaximumSize(size);
		_upperRowHeaderPanel.validate();

		_upperViewportPanel.setPreferredSize(size);
		_upperViewportPanel.setSize(size);
		_upperViewportPanel.setMinimumSize(size);
		_upperViewportPanel.setMaximumSize(size);
		_upperViewportPanel.validate();
	}

	/**
	 * Setzt fr die unteren Platzhalter des Zeilenheaders und des Viewports die Hhe.
	 *
	 * @param height die neue Hhe der Platzhalter
	 */
	private void setLowerPanel(final int height) {
		Dimension size = new Dimension(1, height);
		_lowerRowHeaderPanel.setPreferredSize(size);
		_lowerRowHeaderPanel.setSize(size);
		_lowerRowHeaderPanel.setMinimumSize(size);
		_lowerRowHeaderPanel.setMaximumSize(size);
		_lowerRowHeaderPanel.validate();

		_lowerViewportPanel.setPreferredSize(size);
		_lowerViewportPanel.setSize(size);
		_lowerViewportPanel.setMinimumSize(size);
		_lowerViewportPanel.setMaximumSize(size);
		_lowerViewportPanel.validate();
	}


	/**
	 * Prft, wieviel Platz die Komponente in der Hhe verbraucht und ob sie noch in der ScrollPane angezeigt werden kann.
	 *
	 * @return gibt an, ob sie noch in der ScrollPane angezeigt werden kann
	 */
	private boolean checkRowHeight() {
		_screenHeight = _verticalScrollBar.getVisibleAmount();
		Dataset dataset = (Dataset)_datasets.get(0);
		int visibleAmount = _screenHeight + dataset.getHeight();

		int usedHeight = _upperViewportPanel.getSize().height + _viewportPanel.getSize().height;
		usedHeight += _lowerViewportPanel.getSize().height;
		if(usedHeight < visibleAmount) {	   // ist dann noch platz -> true, sonst false
			return true;
		}
		else {
			return false;
		}
	}

	/* ############### Klasse VerticalScrollBarAdjustmentListener #################### */

	/**
	 * Wird bentigt, um das Scrollverhalten des vertikalen Scrollbalkens zu steuern. Je nachdem, wohin der Scrollbalken verschoben wird, muss der Viewport
	 * aktualisiert werden, damit nicht zu allen Datenstzen die Komponenten angezeigt werden mssen. Dies spart Speicherplatz.
	 */
	private class VerticalScrollBarAdjustmentListener implements AdjustmentListener {

		/* ############## Variablen ############## */
		/** Gibt an, welcher Datensatz der oberste noch sichtbare Datensatz im Fenster ist. */
		private int _position;

		/** Die Hhe des obersten Platzhalters. */
		private int _upperHeight;

		/** Die Hhe des Viewport im ScrollPane. */
		private int _viewportHeight;

		/** Die aktuelle Position des Schiebers im vertikalen Scrollbalken. */
		private int _value;

		/** Die Hhe des unteren Platzhalters. */
		private int _lowerHeight;

		private final Object _lock = new Object();

		/* ############### Methoden der Klasse ############## */

		/** Ermittelt anhand des AdjustmentEvents den Index des obersten sichtbaren Datensatzes und die Hhe fr den oberen Platzhalter. */
		private void init() {
			_screenHeight = _verticalScrollBar.getVisibleAmount();
			_position = 0;
			_upperHeight = 0;
			Dataset dataset = null;
			while(_upperHeight <= _value) {
//                Object obj = _datasets.get(_position);
//                System.out.println("Objekt: " + obj.toString());
				dataset = (Dataset)_datasets.get(_position++);
				_upperHeight += dataset.getHeight();
			}
			if(dataset != null) {
				_position--;
				_upperHeight -= dataset.getHeight();
			}
			_viewportHeight = _viewportPanel.getSize().height;
			_lowerHeight = _lowerViewportPanel.getSize().height;
		}

		/**
		 * Hier wird geprft, welche Datenstze gerade im sichtbaren Bereich des ScrollPane angezeigt werden knnen. Datenstze, die fehlen, werden hinzugefgt,
		 * Datenstze die nicht mehr bentigt werden, werden aus der Ansicht entfernt.
		 *
		 * @param e Event vom Scrollbalken
		 */
		public void adjustmentValueChanged(AdjustmentEvent e) {
			synchronized(_lock) {
				// falls es keine Daten gibt zum Darstellen ->  mache nichts
				if(_datasets.isEmpty()) return;
				_value = e.getValue();

				init();

//            System.out.println("Neue Position: " + _position + " \tValue: " + _value);

				/* ########## */
				if(_position >= _firstRow && _position <= _lastRow) {  // langsam nach unten scrollen
//                System.out.println("erst lschen, dann malen");

					while(_firstRow < _position) {
						Dataset nextDataset = (Dataset)_datasets.get(_firstRow);
						_viewportHeight -= nextDataset.getHeight();
						nextDataset.unsetLinks();
						_rowHeaderPanel.remove(0);
						_viewportPanel.remove(0);
						_firstRow++;
					}
					setUpperPanel(_upperHeight);
					_firstRow = _position;
					Dataset firstDataset = (Dataset)_datasets.get(_firstRow);
					int firstRowHeight = firstDataset.getHeight();
					_lastRow++;

					while((_upperHeight + _viewportHeight < _value + _screenHeight + firstRowHeight) && (_lastRow < _datasets.size())) {
						Dataset lastDataset = (Dataset)_datasets.get(_lastRow++);
						lastDataset.setLinks();

						JComponent rowHeaderRow = lastDataset.getRowHeaderRow(TIME_FORMAT);
						JComponent viewportRow = lastDataset.getViewportRow();

						int lastRowHeight = lastDataset.getHeight();
						_viewportHeight += lastRowHeight;
						_lowerHeight -= lastRowHeight;

						rowHeaderRow.setPreferredSize(new Dimension(rowHeaderRow.getPreferredSize().width, lastRowHeight));
						rowHeaderRow.setMaximumSize(new Dimension(rowHeaderRow.getMaximumSize().width, lastRowHeight));
						viewportRow.setMaximumSize(new Dimension(viewportRow.getMaximumSize().width, lastRowHeight));

						_rowHeaderPanel.add(rowHeaderRow);
						_viewportPanel.add(viewportRow);
					}
					_lastRow--;

					_rowHeaderPanel.revalidate();
					_viewportPanel.revalidate();

					// untere Platzhalter bestimmen
					setLowerPanel(_lowerHeight);
					int anzahl = _viewportPanel.getComponentCount();
					if(anzahl != ((_lastRow - _firstRow) + 1)) {
						_debug.error(
								"Anzahl der Komponenten (" + _viewportPanel.getComponentCount() + ") stimmt nicht mit _firstRow (" + _firstRow
								+ ") und _lastRow (" + _lastRow + ") berein!"
						);
					}
//				System.out.println("Header wird validiert (langsam nach unten).");
//				_headerScrollPane.validate();
				}
				else

					/* ############ */
					if(_position < _firstRow) {	// nach oben scrollen
//                System.out.println("erst malen, dann lschen");
						setUpperPanel(_upperHeight);

						int newFirstRow = _position;
						int newLastRow = _position;

						Dataset firstDataset = (Dataset)_datasets.get(newLastRow);
						int firstRowHeight = firstDataset.getHeight();

						_viewportHeight = 0;
						int index = 0;
						while(_upperHeight + _viewportHeight < _value + _screenHeight + firstRowHeight && newLastRow <= _lastRow) {
							Dataset lastDataset = (Dataset)_datasets.get(newLastRow);
							if(newLastRow < _firstRow) {   // neue Zeilen hinzufgen
								lastDataset.setLinks();

								JComponent rowHeaderRow = lastDataset.getRowHeaderRow(TIME_FORMAT);
								JComponent viewportRow = lastDataset.getViewportRow();

								int lastRowHeight = lastDataset.getHeight();
								_viewportHeight += lastRowHeight;

								rowHeaderRow.setPreferredSize(new Dimension(rowHeaderRow.getPreferredSize().width, lastRowHeight));
								rowHeaderRow.setMaximumSize(new Dimension(rowHeaderRow.getMaximumSize().width, lastRowHeight));
								viewportRow.setMaximumSize(new Dimension(viewportRow.getMaximumSize().width, lastRowHeight));

								_rowHeaderPanel.add(rowHeaderRow, index);
								_viewportPanel.add(viewportRow, index);
							}
							else {
								_viewportHeight += lastDataset.getHeight();
							}

							newLastRow++;
							index++;
						}
						newLastRow--;   // _lastRow
						index--;

						// letzte Elemente lschen
						if(newLastRow < _firstRow) {
							// alles zwischen _firstRow und _lastRow lschen
							for(int i = _lastRow; i >= _firstRow; i--) {
								Dataset dataset = (Dataset)_datasets.get(i);
								dataset.unsetLinks();
								_rowHeaderPanel.remove(_rowHeaderPanel.getComponentCount() - 1);
								_viewportPanel.remove(_viewportPanel.getComponentCount() - 1);
							}
						}
						if(newLastRow >= _firstRow) { // letzten Element lschen
							for(int i = _lastRow; i > newLastRow; i--) {
								Dataset dataset = (Dataset)_datasets.get(i);
								dataset.unsetLinks();
								_rowHeaderPanel.remove(_rowHeaderPanel.getComponentCount() - 1);
								_viewportPanel.remove(_viewportPanel.getComponentCount() - 1);
							}
						}
						_firstRow = newFirstRow;
						_lastRow = newLastRow;

						_rowHeaderPanel.validate();
						_viewportPanel.validate();

						// untere Platzhalter
						_lowerHeight = 0;
						for(int i = _lastRow + 1, n = _datasets.size(); i < n; i++) {
							Dataset dataset = (Dataset)_datasets.get(i);
							_lowerHeight += dataset.getHeight();
						}
						setLowerPanel(_lowerHeight);
						int anzahl = _viewportPanel.getComponentCount();
						if(anzahl != (_lastRow - _firstRow + 1)) {
							_debug.error("Anzahl der Komponenten stimmt nicht mit _firstRow und _lastRow berein!");
						}
//				System.out.println("Header wird validiert (nach oben).");
//				_headerScrollPane.validate();
					}
					else

						/* ############ */
						if(_position > _lastRow) {
//                System.out.println("erst alles lschen, dann malen");
							while(_firstRow <= _lastRow) {
								Dataset nextDataset = (Dataset)_datasets.get(_firstRow);
								_viewportHeight -= nextDataset.getHeight();
								nextDataset.unsetLinks();
								_rowHeaderPanel.remove(0);
								_viewportPanel.remove(0);
								_firstRow++;
							}
							_firstRow = _position;
							_lastRow = _position;

							// oberen Platzhalter ermitteln
							setUpperPanel(_upperHeight);

							Dataset firstDataset = (Dataset)_datasets.get(_firstRow);
							int firstRowHeight = firstDataset.getHeight();
							while((_upperHeight + _viewportHeight < _value + _screenHeight + firstRowHeight) && (_lastRow < _datasets.size())) {
								Dataset lastDataset = (Dataset)_datasets.get(_lastRow++);
								lastDataset.setLinks();

								JComponent rowHeaderRow = lastDataset.getRowHeaderRow(TIME_FORMAT);
								JComponent viewportRow = lastDataset.getViewportRow();

								int lastRowHeight = lastDataset.getHeight();
								_viewportHeight += lastRowHeight;

								rowHeaderRow.setPreferredSize(new Dimension(rowHeaderRow.getPreferredSize().width, lastRowHeight));
								rowHeaderRow.setMaximumSize(new Dimension(rowHeaderRow.getMaximumSize().width, lastRowHeight));
								viewportRow.setMaximumSize(new Dimension(viewportRow.getMaximumSize().width, lastRowHeight));

								_rowHeaderPanel.add(rowHeaderRow);
								_viewportPanel.add(viewportRow);

								_rowHeaderPanel.validate();
								_viewportPanel.validate();
							}

							_lastRow--;
							// LowerHeight ermitteln
							_lowerHeight = 0;
							for(int i = _lastRow + 1, n = _datasets.size(); i < n; i++) {
								Dataset dataset = (Dataset)_datasets.get(i);
								_lowerHeight += dataset.getHeight();
							}
							setLowerPanel(_lowerHeight);
							int anzahl = _viewportPanel.getComponentCount();
							if(anzahl != (_lastRow - _firstRow + 1)) {
								_debug.error("Anzahl der Komponenten stimmt nicht mit _firstRow und _lastRow berein!");
							}
//				System.out.println("Header wird validiert (nach unten).");
//				_headerScrollPane.validate();
						}
				_headerScrollPane.validate();
//            _debug.finest("Datenstze von " + _firstRow + " bis " + _lastRow + " werden angezeigt!");
//            _debug.finest("Anzahl der Komponenten: " + _viewportPanel.getComponentCount());
			}
		}
	}
}
