/*
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * Copyright 2008 by Kappich Systemberatung, Aachen
 * Copyright 2006 by Kappich Systemberatung Aachen
 * Copyright 2006 by Kappich+Kni Systemberatung Aachen (K2S)
 * 
 * This file is part of de.bsvrz.puk.config.
 * 
 * de.bsvrz.puk.config is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.puk.config is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.puk.config; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.puk.config.configFile.datamodel;

import de.bsvrz.puk.config.configFile.fileaccess.SystemObjectInformationInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.config.ConfigurationArea;
import de.bsvrz.dav.daf.main.config.DynamicObject;
import de.bsvrz.dav.daf.main.config.DynamicObjectType;
import de.bsvrz.dav.daf.main.config.InvalidationListener;
import de.bsvrz.dav.daf.main.config.SystemObjectType;
import de.bsvrz.dav.daf.main.config.MutableCollectionChangeListener;
import de.bsvrz.dav.daf.main.config.SystemObject;

import java.util.*;

/**
 * Implementierung des Interfaces fr den Typ von dynamischen Objekten.
 *
 * @author Kappich Systemberatung
 * @version $Revision: 6393 $
 */
public class ConfigDynamicObjectType extends ConfigSystemObjectType implements DynamicObjectType {

	/** Objekt zur Verwaltung von Anmeldungen auf nderung der Elemente dieses Typs. */
	private ConfigMutableCollectionSupport _mutableCollectionSupport = new ConfigMutableCollectionSupport(this);

	/** Enthlt alle Listener, die informiert werden mssen sobald ein Objekt dieses Typs erzeugt wurde. */
	private final Set<DynamicObjectCreatedListener> _createdListener = new HashSet<DynamicObjectCreatedListener>();

	/** Alle Listener, die informiert werden sobald sich der Name eines dynamischen Objekt ndert */
	private final Set<NameChangeListener> _nameChangedListener = new HashSet<NameChangeListener>();

	/** Alle Listener, die informiert werden wollen sobald ein dynamisches Objekt ungltig wird. */
	private final Set<InvalidationListener> _invalidationListener = new HashSet<InvalidationListener>();

	/**
	 * Konstruktor erstellt den Typ eines dynamischen Objekts.
	 *
	 * @param configurationArea der Konfigurationsbereich des Typs
	 * @param systemObjectInfo  das korrespondierende Objekt fr die Dateioperationen des Typ-Objekts
	 */
	public ConfigDynamicObjectType(ConfigurationArea configurationArea, SystemObjectInformationInterface systemObjectInfo) {
		super(configurationArea, systemObjectInfo);
	}

	public PersistenceMode getPersistenceMode() {
		Data data = getConfigurationData(getDataModel().getAttributeGroup("atg.dynamischerTypEigenschaften"), getDataModel().getAspect("asp.eigenschaften"));
		if(data != null) {
			int mode = data.getUnscaledValue("persistenzModus").intValue();
			switch(mode) {
				case 1:
					return PersistenceMode.TRANSIENT_OBJECTS;
				case 2:
					return PersistenceMode.PERSISTENT_OBJECTS;
				case 3:
					return PersistenceMode.PERSISTENT_AND_INVALID_ON_RESTART;
				default:
					throw new IllegalStateException("Persistenzmodus am '" + getPid() + "' hat einen ungltigen Wert: " + mode);
			}
		}
		else {
			throw new IllegalStateException("Persistenzmodus am '" + getPid() + "' kann nicht ermittelt werden.");
		}
	}

	public void addObjectCreationListener(DynamicObjectCreatedListener objectCreatedListener) {
		synchronized(_createdListener) {
			_createdListener.add(objectCreatedListener);
		}
	}

	public void removeObjectCreationListener(DynamicObjectCreatedListener objectCreatedListener) {
		synchronized(_createdListener) {
			_createdListener.add(objectCreatedListener);
		}
	}

	public void addNameChangeListener(NameChangeListener nameChangeListener) {
		synchronized(_nameChangedListener) {
			_nameChangedListener.add(nameChangeListener);
		}
	}

	public void removeNameChangeListener(NameChangeListener nameChangeListener) {
		synchronized(_nameChangedListener) {
			_nameChangedListener.remove(nameChangeListener);
		}
	}

	public void addInvalidationListener(InvalidationListener invalidationListener) {
		synchronized(_invalidationListener) {
			_invalidationListener.add(invalidationListener);
		}
	}

	public void removeInvalidationListener(InvalidationListener invalidationListener) {
		synchronized(_invalidationListener) {
			_invalidationListener.remove(invalidationListener);
		}
	}

	/**
	 * Informiert alle Listener, dass ein dynamischens Objekt ungltig geworden ist. Ist kein Listener angemeldet, wird nichts gemacht.
	 *
	 * @param newInvalidObject Objekt, das ungltig geworden ist.
	 */
	public void informInvalidationListener(DynamicObject newInvalidObject) {
		final boolean isLocalDynamicObject = newInvalidObject instanceof ConfigDynamicObject;
		if(isLocalDynamicObject) handleDeletedElement(newInvalidObject);
		synchronized(_invalidationListener) {
			for(InvalidationListener invalidationListener : _invalidationListener) {
				invalidationListener.invalidObject(newInvalidObject);
			}
			if(isLocalDynamicObject) {
				// Alle Supertypen benachrichtigen
				final List<SystemObjectType> superTyps = getSuperTypes();
				for(SystemObjectType superTyp : superTyps) {
					((ConfigDynamicObjectType)superTyp).informInvalidationListener(newInvalidObject);
				}
			}
		}
	}

	/**
	 * Informiert alle Listener, die sich auf Namennderungen angemeldet haben. Ist kein Listener vorhanden wird nichts gemacht.
	 *
	 * @param newNamedObject Objekt, dessen Name gendert wurde.
	 */
	public void informNameChangedListener(DynamicObject newNamedObject) {
		final boolean isLocalDynamicObject = newNamedObject instanceof ConfigDynamicObject;
		synchronized(_nameChangedListener) {

			for(NameChangeListener nameChangeListener : _nameChangedListener) {
				nameChangeListener.nameChanged(newNamedObject);
			}

			if(isLocalDynamicObject) {
				// Alle Supertypen benachrichtigen
				final List<SystemObjectType> superTyps = getSuperTypes();
				for(SystemObjectType superTyp : superTyps) {
					((ConfigDynamicObjectType)superTyp).informNameChangedListener(newNamedObject);
				}
			}
		}
	}

	/**
	 * Diese Methode wird aufgerufen, wenn ein neues Objekt angelegt wurde und informiert alle Listener, die auf nderungen dieser Art angemeldet sind. Ist kein
	 * Listener vorhanden wird nichts gemacht.
	 *
	 * @param createdObject Objekt, das neu erzeugt wurde.
	 * @param simulationVariant Simulationsvariante unter der das Objekt erzeugt wurde.
	 */
	public void informCreateListener(DynamicObject createdObject, final short simulationVariant) {
		handleAddedElement(createdObject, simulationVariant);
		synchronized(_createdListener) {
			for(DynamicObjectCreatedListener dynamicObjectCreatedListener : _createdListener) {
				dynamicObjectCreatedListener.objectCreated(createdObject);
			}
			// Alle Supertypen benachrichtigen
			final List<SystemObjectType> superTyps = getSuperTypes();
			for(SystemObjectType superTyp : superTyps) {
				((ConfigDynamicObjectType)superTyp).informCreateListener(createdObject, simulationVariant);
			}
		}
	}

	/**
	 * Speichert neue Objekte dieses Typs und stt die Benachrichtigung der Listener an, die sich auf nderungen der Elemente angemeldet haben.
	 * @param createdObject Neu erzeugtes Objekt
	 * @param simulationVariant Simulationsvariante unter der das Objekt erzeugt wurde.
	 */
	private void handleAddedElement(final DynamicObject createdObject, final short simulationVariant) {
		getAllElements().add(createdObject);
		final List<SystemObject> addedElements = new ArrayList<SystemObject>(1);
		addedElements.add(createdObject);
		final List<SystemObject> removedElements = new ArrayList<SystemObject>();
		((ConfigDataModel)getDataModel()).sendCollectionChangedNotification(_mutableCollectionSupport, simulationVariant, addedElements, removedElements);
	}

	/**
	 * Entfernt ein gelschtes Element dieses Typs und stt die Benachrichtigung der Listener an, die sich auf nderungen der Elemente angemeldet haben.
	 * @param invalidatedObject gelschtes Objekt
	 */
	public void handleDeletedElement(final DynamicObject invalidatedObject) {
		getAllElements().remove(invalidatedObject);
		final List<SystemObject> addedElements = new ArrayList<SystemObject>(0);
		final List<SystemObject> removedElements = new ArrayList<SystemObject>(1);
		removedElements.add(invalidatedObject);
		((ConfigDataModel)getDataModel()).sendCollectionChangedNotification(_mutableCollectionSupport,
		                                                                    ((ConfigDynamicObject)invalidatedObject).getSimulationVariant(), addedElements, removedElements);
	}

	public void addChangeListener(short simulationVariant, MutableCollectionChangeListener changeListener) {
		getAllElements();
		_mutableCollectionSupport.addChangeListener(simulationVariant, changeListener);
	}

	public void removeChangeListener(short simulationVariant, MutableCollectionChangeListener changeListener) {
		_mutableCollectionSupport.removeChangeListener(simulationVariant, changeListener);
	}

	public List<SystemObject> getElements(short simulationVariant) {
		final List<SystemObject> allElements = getAllElements();
		final List<SystemObject> elements = new ArrayList<SystemObject>();
		for(SystemObject element : allElements) {
			if(element instanceof ConfigDynamicObject) {
				ConfigDynamicObject configDynamicObject = (ConfigDynamicObject)element;
				if(configDynamicObject.getSimulationVariant() == simulationVariant) elements.add(element);
			}
		}
		return elements;
	}

}
